@@ -5,7 +5,7 @@
  * Originally written by Tatsuo Ishii and enhanced by many contributors.
  *
  * contrib/pgbench/pgbench.c
- * Copyright (c) 2000-2011, PostgreSQL Global Development Group
+ * Copyright (c) 2000-2012, PostgreSQL Global Development Group
  * ALL RIGHTS RESERVED;
  *
  * Permission to use, copy, modify, and distribute this software and its
@@ -33,6 +33,7 @@
 
 #include "postgres_fe.h"
 
+#include "getopt_long.h"
 #include "libpq-fe.h"
 #include "libpq/pqsignal.h"
 #include "portability/instr_time.h"
@@ -44,10 +45,6 @@
 #include <unistd.h>
 #endif   /* ! WIN32 */
 
-#ifdef HAVE_GETOPT_H
-#include <getopt.h>
-#endif
-
 #ifdef HAVE_SYS_SELECT_H
 #include <sys/select.h>
 #endif
@@ -69,7 +66,7 @@
 typedef struct win32_pthread *pthread_t;
 typedef int pthread_attr_t;
 
-static int	pthread_create(pthread_t *thread, pthread_attr_t * attr, void *(*start_routine) (void *), void *arg);
+static int	pthread_create(pthread_t *thread, pthread_attr_t *attr, void *(*start_routine) (void *), void *arg);
 static int	pthread_join(pthread_t th, void **thread_return);
 #elif defined(ENABLE_THREAD_SAFETY)
 /* Use platform-dependent pthread capability */
@@ -87,7 +84,7 @@ static int	pthread_join(pthread_t th, void **thread_return);
 typedef struct fork_pthread *pthread_t;
 typedef int pthread_attr_t;
 
-static int	pthread_create(pthread_t *thread, pthread_attr_t * attr, void *(*start_routine) (void *), void *arg);
+static int	pthread_create(pthread_t *thread, pthread_attr_t *attr, void *(*start_routine) (void *), void *arg);
 static int	pthread_join(pthread_t th, void **thread_return);
 #endif
 
@@ -122,6 +119,17 @@ int			scale = 1;
  */
 int			fillfactor = 100;
 
+/*
+ * use unlogged tables?
+ */
+int			unlogged_tables = 0;
+
+/*
+ * tablespace selection
+ */
+char	   *tablespace = NULL;
+char	   *index_tablespace = NULL;
+
 /*
  * end of configurable parameters
  *********************************************************************/
@@ -132,7 +140,7 @@ int			fillfactor = 100;
 #define naccounts	100000
 
 #ifdef PGXC
-bool		use_branch = true;	/* use branch id in DDL and DML */
+bool		use_branch = false;	/* use branch id in DDL and DML */
 #endif
 bool		use_log;			/* log transaction latencies to a file */
 bool		is_connect;			/* establish connection for each transaction */
@@ -193,6 +201,7 @@ typedef struct
 	instr_time	start_time;		/* thread start time */
 	instr_time *exec_elapsed;	/* time spent executing cmds (per Command) */
 	int		   *exec_count;		/* number of cmd executions (per Command) */
+	unsigned short random_state[3];		/* separate randomness for each thread */
 } TState;
 
 #define INVALID_THREAD		((pthread_t) 0)
@@ -255,17 +264,17 @@ static char *tpc_b = {
 
 #ifdef PGXC
 static char *tpc_b_bid = {
-	"\\set nbranches :scale\n"
-	"\\set ntellers 10 * :scale\n"
-	"\\set naccounts 100000 * :scale\n"
+	"\\set nbranches " CppAsString2(nbranches) " * :scale\n"
+	"\\set ntellers " CppAsString2(ntellers) " * :scale\n"
+	"\\set naccounts " CppAsString2(naccounts) " * :scale\n"
 	"\\setrandom aid 1 :naccounts\n"
 	"\\setrandom bid 1 :nbranches\n"
 	"\\setrandom tid 1 :ntellers\n"
 	"\\setrandom delta -5000 5000\n"
 	"BEGIN;\n"
-	"UPDATE pgbench_accounts SET abalance = abalance + :delta WHERE aid = :aid;\n"
-	"SELECT abalance FROM pgbench_accounts WHERE aid = :aid\n"
-	"UPDATE pgbench_tellers SET tbalance = tbalance + :delta WHERE tid = :tid;\n"
+	"UPDATE pgbench_accounts SET abalance = abalance + :delta WHERE aid = :aid AND bid = :bid;\n"
+	"SELECT abalance FROM pgbench_accounts WHERE aid = :aid AND bid = :bid\n"
+	"UPDATE pgbench_tellers SET tbalance = tbalance + :delta WHERE tid = :tid AND bid = :bid;\n"
 	"UPDATE pgbench_branches SET bbalance = bbalance + :delta WHERE bid = :bid;\n"
 	"INSERT INTO pgbench_history (tid, bid, aid, delta, mtime) VALUES (:tid, :bid, :aid, :delta, CURRENT_TIMESTAMP);\n"
 	"END;\n"
@@ -291,9 +300,9 @@ static char *simple_update = {
 
 #ifdef PGXC
 static char *simple_update_bid = {
-	"\\set nbranches :scale\n"
-	"\\set ntellers 10 * :scale\n"
-	"\\set naccounts 100000 * :scale\n"
+	"\\set nbranches " CppAsString2(nbranches) " * :scale\n"
+	"\\set ntellers " CppAsString2(ntellers) " * :scale\n"
+	"\\set naccounts " CppAsString2(naccounts) " * :scale\n"
 	"\\setrandom aid 1 :naccounts\n"
 	"\\setrandom bid 1 :nbranches\n"
 	"\\setrandom tid 1 :ntellers\n"
@@ -326,6 +335,9 @@ xmalloc(size_t size)
 {
 	void	   *result;
 
+	/* Avoid unportable behavior of malloc(0) */
+	if (size == 0)
+		size = 1;
 	result = malloc(size);
 	if (!result)
 	{
@@ -340,6 +352,9 @@ xrealloc(void *ptr, size_t size)
 {
 	void	   *result;
 
+	/* Avoid unportable behavior of realloc(NULL, 0) */
+	if (ptr == NULL && size == 0)
+		size = 1;
 	result = realloc(ptr, size);
 	if (!result)
 	{
@@ -369,14 +384,20 @@ usage(const char *progname)
 {
 	printf("%s is a benchmarking tool for PostgreSQL.\n\n"
 		   "Usage:\n"
-		   "  %s [OPTIONS]... [DBNAME]\n"
+		   "  %s [OPTION]... [DBNAME]\n"
 		   "\nInitialization options:\n"
 		   "  -i           invokes initialization mode\n"
 		   "  -F NUM       fill factor\n"
 #ifdef PGXC
-		   "  -k           distribute by primary key (default: branch id - bid)\n"
+		   "  -k           distribute by primary key branch id - bid\n"
 #endif
 		   "  -s NUM       scaling factor\n"
+		   "  --index-tablespace=TABLESPACE\n"
+		   "               create indexes in the specified tablespace\n"
+		   "  --tablespace=TABLESPACE\n"
+		   "               create tables in the specified tablespace\n"
+		   "  --unlogged-tables\n"
+		   "               create tables as unlogged tables\n"
 		   "\nBenchmarking options:\n"
 		"  -c NUM       number of concurrent database clients (default: 1)\n"
 		   "  -C           establish new connection for each transaction\n"
@@ -384,11 +405,11 @@ usage(const char *progname)
 		   "               define variable for use by custom script\n"
 		   "  -f FILENAME  read transaction script from FILENAME\n"
 #ifdef PGXC
-		   "  -k           query with key only, no branch id (bid)\n"
+		   "  -k           query with default key and additional key branch id (bid)\n"
 #endif
 		   "  -j NUM       number of threads (default: 1)\n"
 		   "  -l           write transaction times to log file\n"
-		   "  -M {simple|extended|prepared}\n"
+		   "  -M simple|extended|prepared\n"
 		   "               protocol for submitting queries to server (default: simple)\n"
 		   "  -n           do not run VACUUM before tests\n"
 		   "  -N           do not update tables \"pgbench_tellers\" and \"pgbench_branches\"\n"
@@ -399,12 +420,12 @@ usage(const char *progname)
 		   "  -T NUM       duration of benchmark test in seconds\n"
 		   "  -v           vacuum all four standard tables before tests\n"
 		   "\nCommon options:\n"
-		   "  -d           print debugging output\n"
-		   "  -h HOSTNAME  database server host or socket directory\n"
-		   "  -p PORT      database server port number\n"
-		   "  -U USERNAME  connect as specified database user\n"
-		   "  --help       show this help, then exit\n"
-		   "  --version    output version information, then exit\n"
+		   "  -d             print debugging output\n"
+		   "  -h HOSTNAME    database server host or socket directory\n"
+		   "  -p PORT        database server port number\n"
+		   "  -U USERNAME    connect as specified database user\n"
+		   "  -V, --version  output version information, then exit\n"
+		   "  -?, --help     show this help, then exit\n"
 		   "\n"
 		   "Report bugs to <pgsql-bugs@postgresql.org>.\n",
 		   progname, progname);
@@ -412,13 +433,18 @@ usage(const char *progname)
 
 /* random number generator: uniform distribution from min to max inclusive */
 static int
-getrand(int min, int max)
+getrand(TState *thread, int min, int max)
 {
 	/*
 	 * Odd coding is so that min and max have approximately the same chance of
 	 * being selected as do numbers between them.
+	 *
+	 * pg_erand48() is thread-safe and concurrent, which is why we use it
+	 * rather than random(), which in glibc is non-reentrant, and therefore
+	 * protected by a mutex, and therefore a bottleneck on machines with many
+	 * CPUs.
 	 */
-	return min + (int) (((max - min + 1) * (double) random()) / (MAX_RANDOM_VALUE + 1.0));
+	return min + (int) ((max - min + 1) * pg_erand48(thread->random_state));
 }
 
 /* call PQexec() and exit() on failure */
@@ -933,7 +959,7 @@ top:
 		if (commands[st->state] == NULL)
 		{
 			st->state = 0;
-			st->use_file = getrand(0, num_files - 1);
+			st->use_file = getrand(thread, 0, num_files - 1);
 			commands = sql_files[st->use_file];
 		}
 	}
@@ -1092,17 +1118,31 @@ top:
 			else
 				max = atoi(argv[3]);
 
-			if (max < min || max > MAX_RANDOM_VALUE)
+			if (max < min)
 			{
-				fprintf(stderr, "%s: invalid maximum number %d\n", argv[0], max);
+				fprintf(stderr, "%s: maximum is less than minimum\n", argv[0]);
+				st->ecnt++;
+				return true;
+			}
+
+			/*
+			 * getrand() neeeds to be able to subtract max from min and add
+			 * one the result without overflowing.	Since we know max > min,
+			 * we can detect overflow just by checking for a negative result.
+			 * But we must check both that the subtraction doesn't overflow,
+			 * and that adding one to the result doesn't overflow either.
+			 */
+			if (max - min < 0 || (max - min) + 1 < 0)
+			{
+				fprintf(stderr, "%s: range too large\n", argv[0]);
 				st->ecnt++;
 				return true;
 			}
 
 #ifdef DEBUG
-			printf("min: %d max: %d random: %d\n", min, max, getrand(min, max));
+			printf("min: %d max: %d random: %d\n", min, max, getrand(thread, min, max));
 #endif
-			snprintf(res, sizeof(res), "%d", getrand(min, max));
+			snprintf(res, sizeof(res), "%d", getrand(thread, min, max));
 
 			if (!putVariable(st, argv[0], argv[1], res))
 			{
@@ -1279,41 +1319,49 @@ init(void)
 	 * versions.  Since pgbench has never pretended to be fully TPC-B
 	 * compliant anyway, we stick with the historical behavior.
 	 */
-	static char *DDLs[] = {
-		"drop table if exists pgbench_branches",
+	struct ddlinfo
+	{
+		char	   *table;
+		char	   *cols;
+		int			declare_fillfactor;
 #ifdef PGXC
-		/* use primary key as distribution column */
-		"create table pgbench_branches(bid int not null,bbalance int,filler char(88)) with (fillfactor=%d) distribute by (bid)",
-		"drop table if exists pgbench_tellers",
-		"create table pgbench_tellers(tid int not null,bid int,tbalance int,filler char(84)) with (fillfactor=%d) distribute by (tid)",
-		"drop table if exists pgbench_accounts",
-		"create table pgbench_accounts(aid int not null,bid int,abalance int,filler char(84)) with (fillfactor=%d) distribute by (aid)",
-		"drop table if exists pgbench_history",
-		"create table pgbench_history(tid int,bid int,aid int,delta int,mtime timestamp,filler char(22)) distribute by (bid)"
-#else
-		"create table pgbench_branches(bid int not null,bbalance int,filler char(88)) with (fillfactor=%d)",
-		"drop table if exists pgbench_tellers",
-		"create table pgbench_tellers(tid int not null,bid int,tbalance int,filler char(84)) with (fillfactor=%d)",
-		"drop table if exists pgbench_accounts",
-		"create table pgbench_accounts(aid int not null,bid int,abalance int,filler char(84)) with (fillfactor=%d)",
-		"drop table if exists pgbench_history",
-		"create table pgbench_history(tid int,bid int,aid int,delta int,mtime timestamp,filler char(22))"
+		char	   *distribute_by;
 #endif
 	};
-
+	struct ddlinfo DDLs[] = {
+		{
+			"pgbench_branches",
+			"bid int not null,bbalance int,filler char(88)",
+			1
 #ifdef PGXC
-	/* a version for using bid as distribution column */
-	static char *DDLs_bid[] = {
-		"drop table if exists pgbench_branches",
-		"create table pgbench_branches(bid int not null,bbalance int,filler char(88)) with (fillfactor=%d) distribute by (bid)",
-		"drop table if exists pgbench_tellers",
-		"create table pgbench_tellers(tid int not null,bid int,tbalance int,filler char(84)) with (fillfactor=%d) distribute by (bid)",
-		"drop table if exists pgbench_accounts",
-		"create table pgbench_accounts(aid int not null,bid int,abalance int,filler char(84)) with (fillfactor=%d) distribute by (bid)",
-		"drop table if exists pgbench_history",
-		"create table pgbench_history(tid int,bid int,aid int,delta int,mtime timestamp,filler char(22)) distribute by (bid)"
-	};
+			, "distribute by hash (bid)"
 #endif
+		},
+		{
+			"pgbench_tellers",
+			"tid int not null,bid int,tbalance int,filler char(84)",
+			1
+#ifdef PGXC
+			, "distribute by hash (bid)"
+#endif
+		},
+		{
+			"pgbench_accounts",
+			"aid int not null,bid int,abalance int,filler char(84)",
+			1
+#ifdef PGXC
+			, "distribute by hash (bid)"
+#endif
+		},
+		{
+			"pgbench_history",
+			"tid int,bid int,aid int,delta int,mtime timestamp,filler char(22)",
+			0
+#ifdef PGXC
+			, "distribute by hash (bid)"
+#endif
+		}
+	};
 
 	static char *DDLAFTERs[] = {
 		"alter table pgbench_branches add primary key (bid)",
@@ -1324,10 +1372,10 @@ init(void)
 #ifdef PGXC
 	static char *DDLAFTERs_bid[] = {
 		"alter table pgbench_branches add primary key (bid)",
-		"create index tellers_idx on pgbench_tellers (tid)",
-		"create index accounts_idx on pgbench_accounts (aid)",
-#endif
+		"alter table pgbench_tellers add primary key (tid,bid)",
+		"alter table pgbench_accounts add primary key (aid,bid)"
 	};
+#endif
 
 	PGconn	   *con;
 	PGresult   *res;
@@ -1337,52 +1385,45 @@ init(void)
 	if ((con = doConnect()) == NULL)
 		exit(1);
 
-#ifdef PGXC
-	if (use_branch)
-	{
-		for (i = 0; i < lengthof(DDLs_bid); i++)
-		{
-			/*
-			 * set fillfactor for branches, tellers and accounts tables
-			 */
-			if ((strstr(DDLs_bid[i], "create table pgbench_branches") == DDLs_bid[i]) ||
-				(strstr(DDLs_bid[i], "create table pgbench_tellers") == DDLs_bid[i]) ||
-				(strstr(DDLs_bid[i], "create table pgbench_accounts") == DDLs_bid[i]))
-			{
-				char		ddl_stmt[128];
-	
-				snprintf(ddl_stmt, 128, DDLs_bid[i], fillfactor);
-				executeStatement(con, ddl_stmt);
-				continue;
-			}
-			else
-				executeStatement(con, DDLs_bid[i]);
-		}
-	} 
-	else
-	{
-#endif
 	for (i = 0; i < lengthof(DDLs); i++)
 	{
-		/*
-		 * set fillfactor for branches, tellers and accounts tables
-		 */
-		if ((strstr(DDLs[i], "create table pgbench_branches") == DDLs[i]) ||
-			(strstr(DDLs[i], "create table pgbench_tellers") == DDLs[i]) ||
-			(strstr(DDLs[i], "create table pgbench_accounts") == DDLs[i]))
-		{
-			char		ddl_stmt[128];
+		char		opts[256];
+		char		buffer[256];
+		struct ddlinfo *ddl = &DDLs[i];
 
-			snprintf(ddl_stmt, 128, DDLs[i], fillfactor);
-			executeStatement(con, ddl_stmt);
-			continue;
+		/* Remove old table, if it exists. */
+		snprintf(buffer, 256, "drop table if exists %s", ddl->table);
+		executeStatement(con, buffer);
+
+		/* Construct new create table statement. */
+		opts[0] = '\0';
+		if (ddl->declare_fillfactor)
+			snprintf(opts + strlen(opts), 256 - strlen(opts),
+					 " with (fillfactor=%d)", fillfactor);
+		if (tablespace != NULL)
+		{
+			char	   *escape_tablespace;
+
+			escape_tablespace = PQescapeIdentifier(con, tablespace,
+												   strlen(tablespace));
+			snprintf(opts + strlen(opts), 256 - strlen(opts),
+					 " tablespace %s", escape_tablespace);
+			PQfreemem(escape_tablespace);
 		}
-		else
-			executeStatement(con, DDLs[i]);
-	}
 #ifdef PGXC
-	}
+		/* Add distribution columns if necessary */
+		if (use_branch)
+			snprintf(buffer, 256, "create%s table %s(%s)%s %s",
+					 unlogged_tables ? " unlogged" : "",
+					 ddl->table, ddl->cols, opts, ddl->distribute_by);
+		else
 #endif
+		snprintf(buffer, 256, "create%s table %s(%s)%s",
+				 unlogged_tables ? " unlogged" : "",
+				 ddl->table, ddl->cols, opts);
+
+		executeStatement(con, buffer);
+	}
 
 	executeStatement(con, "begin");
 
@@ -1446,23 +1487,55 @@ init(void)
 	/*
 	 * create indexes
 	 */
+	fprintf(stderr, "set primary key...\n");
 #ifdef PGXC
+	/*
+	 * If all the tables are distributed according to bid, create an index on it
+	 * instead.
+	 */
 	if (use_branch)
 	{
-		fprintf(stderr, "set primary keys and indexes...\n");
 		for (i = 0; i < lengthof(DDLAFTERs_bid); i++)
-			executeStatement(con, DDLAFTERs_bid[i]);
+		{
+			char		buffer[256];
+
+			strncpy(buffer, DDLAFTERs_bid[i], 256);
+
+			if (index_tablespace != NULL)
+			{
+				char	   *escape_tablespace;
+
+				escape_tablespace = PQescapeIdentifier(con, index_tablespace,
+												   strlen(index_tablespace));
+				snprintf(buffer + strlen(buffer), 256 - strlen(buffer),
+						 " using index tablespace %s", escape_tablespace);
+				PQfreemem(escape_tablespace);
+			}
+
+			executeStatement(con, buffer);
+		}
 	}
 	else
-	{
 #endif
-	fprintf(stderr, "set primary key...\n");
 	for (i = 0; i < lengthof(DDLAFTERs); i++)
-		executeStatement(con, DDLAFTERs[i]);
-#ifdef PGXC
+	{
+		char		buffer[256];
+
+		strncpy(buffer, DDLAFTERs[i], 256);
+
+		if (index_tablespace != NULL)
+		{
+			char	   *escape_tablespace;
+
+			escape_tablespace = PQescapeIdentifier(con, index_tablespace,
+												   strlen(index_tablespace));
+			snprintf(buffer + strlen(buffer), 256 - strlen(buffer),
+					 " using index tablespace %s", escape_tablespace);
+			PQfreemem(escape_tablespace);
+		}
+
+		executeStatement(con, buffer);
 	}
-#endif
-	
 
 	/* vacuum */
 	fprintf(stderr, "vacuum...");
@@ -1622,7 +1695,7 @@ process_commands(char *buf)
 			{
 				if (pg_strcasecmp(my_commands->argv[2], "us") != 0 &&
 					pg_strcasecmp(my_commands->argv[2], "ms") != 0 &&
-					pg_strcasecmp(my_commands->argv[2], "s"))
+					pg_strcasecmp(my_commands->argv[2], "s") != 0)
 				{
 					fprintf(stderr, "%s: unknown time unit '%s' - must be us, ms or s\n",
 							my_commands->argv[0], my_commands->argv[2]);
@@ -1889,6 +1962,7 @@ main(int argc, char **argv)
 	int			do_vacuum_accounts = 0; /* do vacuum accounts before testing? */
 	int			ttype = 0;		/* transaction type. 0: TPC-B, 1: SELECT only,
 								 * 2: skip update of branches and tellers */
+	int			optindex;
 	char	   *filename = NULL;
 	bool		scale_given = false;
 
@@ -1902,6 +1976,13 @@ main(int argc, char **argv)
 
 	int			i;
 
+	static struct option long_options[] = {
+		{"index-tablespace", required_argument, NULL, 3},
+		{"tablespace", required_argument, NULL, 2},
+		{"unlogged-tables", no_argument, &unlogged_tables, 1},
+		{NULL, 0, NULL, 0}
+	};
+
 #ifdef HAVE_GETRLIMIT
 	struct rlimit rlim;
 #endif
@@ -1946,9 +2027,9 @@ main(int argc, char **argv)
 	memset(state, 0, sizeof(CState));
 
 #ifdef PGXC
-	while ((c = getopt(argc, argv, "ih:knvp:dSNc:j:Crs:t:T:U:lf:D:F:M:")) != -1)
+	while ((c = getopt_long(argc, argv, "ih:knvp:dSNc:j:Crs:t:T:U:lf:D:F:M:", long_options, &optindex)) != -1)
 #else
-	while ((c = getopt(argc, argv, "ih:nvp:dSNc:j:Crs:t:T:U:lf:D:F:M:")) != -1)
+	while ((c = getopt_long(argc, argv, "ih:nvp:dSNc:j:Crs:t:T:U:lf:D:F:M:", long_options, &optindex)) != -1)
 #endif
 	{
 		switch (c)
@@ -1958,7 +2039,7 @@ main(int argc, char **argv)
 				break;
 #ifdef PGXC
 			case 'k':
-				use_branch = false;
+				use_branch = true;
 				break;
 #endif
 			case 'h':
@@ -2106,6 +2187,15 @@ main(int argc, char **argv)
 					exit(1);
 				}
 				break;
+			case 0:
+				/* This covers long options which take no argument. */
+				break;
+			case 2:				/* tablespace */
+				tablespace = optarg;
+				break;
+			case 3:				/* index-tablespace */
+				index_tablespace = optarg;
+				break;
 			default:
 				fprintf(stderr, _("Try \"%s --help\" for more information.\n"), progname);
 				exit(1);
@@ -2307,6 +2397,9 @@ main(int argc, char **argv)
 		thread->tid = i;
 		thread->state = &state[nclients / nthreads * i];
 		thread->nstate = nclients / nthreads;
+		thread->random_state[0] = random();
+		thread->random_state[1] = random();
+		thread->random_state[2] = random();
 
 		if (is_latencies)
 		{
@@ -2449,7 +2542,7 @@ threadRun(void *arg)
 		Command   **commands = sql_files[st->use_file];
 		int			prev_ecnt = st->ecnt;
 
-		st->use_file = getrand(0, num_files - 1);
+		st->use_file = getrand(thread, 0, num_files - 1);
 		if (!doCustom(thread, st, &result->conn_time, logfile))
 			remains--;			/* I've aborted */
 
@@ -2613,13 +2706,12 @@ typedef struct fork_pthread
 
 static int
 pthread_create(pthread_t *thread,
-			   pthread_attr_t * attr,
+			   pthread_attr_t *attr,
 			   void *(*start_routine) (void *),
 			   void *arg)
 {
 	fork_pthread *th;
 	void	   *ret;
-	instr_time	start_time;
 
 	th = (fork_pthread *) xmalloc(sizeof(fork_pthread));
 	if (pipe(th->pipes) < 0)
@@ -2648,17 +2740,6 @@ pthread_create(pthread_t *thread,
 	if (duration > 0)
 		setalarm(duration);
 
-	/*
-	 * Set a different random seed in each child process.  Otherwise they all
-	 * inherit the parent's state and generate the same "random" sequence. (In
-	 * the threaded case, the different threads will obtain subsets of the
-	 * output of a single random() sequence, which should be okay for our
-	 * purposes.)
-	 */
-	INSTR_TIME_SET_CURRENT(start_time);
-	srandom(((unsigned int) INSTR_TIME_GET_MICROSEC(start_time)) +
-			((unsigned int) getpid()));
-
 	ret = start_routine(arg);
 	write(th->pipes[1], ret, sizeof(TResult));
 	close(th->pipes[1]);
@@ -2741,7 +2822,7 @@ win32_pthread_run(void *arg)
 
 static int
 pthread_create(pthread_t *thread,
-			   pthread_attr_t * attr,
+			   pthread_attr_t *attr,
 			   void *(*start_routine) (void *),
 			   void *arg)
 {
