@@ -3,8 +3,8 @@
  * varsup.c
  *	  postgres OID & XID variables support routines
  *
- * Copyright (c) 2000-2011, PostgreSQL Global Development Group
- * Portions Copyright (c) 2010-2013, Postgres-XC Development Group
+ * Copyright (c) 2000-2012, PostgreSQL Global Development Group
+ * Portions Copyright (c) 2010-2012 Postgres-XC Development Group
  *
  * IDENTIFICATION
  *	  src/backend/access/transam/varsup.c
@@ -23,13 +23,12 @@
 #include "postmaster/autovacuum.h"
 #include "storage/pmsignal.h"
 #include "storage/proc.h"
-#include "utils/builtins.h"
+#include "utils/syscache.h"
 #ifdef PGXC
 #include "pgxc/pgxc.h"
 #include "access/gtm.h"
 #include "storage/procarray.h"
 #endif
-#include "utils/syscache.h"
 
 
 /* Number of OIDs to prefetch (preallocate) per XLOG write */
@@ -76,7 +75,7 @@ GetForceXidFromGTM(void)
 /*
  * Allocate the next XID for a new transaction or subtransaction.
  *
- * The new XID is also stored into MyProc before returning.
+ * The new XID is also stored into MyPgXact before returning.
  *
  * Note: when this is called, we are actually already inside a valid
  * transaction, since XIDs are now not allocated until the transaction
@@ -104,7 +103,7 @@ GetNewTransactionId(bool isSubXact)
 	if (IsBootstrapProcessingMode())
 	{
 		Assert(!isSubXact);
-		MyProc->xid = BootstrapTransactionId;
+		MyPgXact->xid = BootstrapTransactionId;
 		return BootstrapTransactionId;
 	}
 
@@ -113,14 +112,17 @@ GetNewTransactionId(bool isSubXact)
 		elog(ERROR, "cannot assign TransactionIds during recovery");
 
 #ifdef PGXC
-	if (IS_PGXC_COORDINATOR && !IsConnFromCoord() || IsPGXCNodeXactDatanodeDirect())
+	/* Initialize transaction ID */
+	xid = InvalidTransactionId;
+
+	if ((IS_PGXC_COORDINATOR && !IsConnFromCoord()) || IsPGXCNodeXactDatanodeDirect())
 	{
 		/*
 		 * Get XID from GTM before acquiring the lock as concurrent connections are
 		 * being handled on GTM side even if the lock is acquired in a different
 		 * order.
 		 */
-		if (IsAutoVacuumWorkerProcess() && (MyProc->vacuumFlags & PROC_IN_VACUUM))
+		if (IsAutoVacuumWorkerProcess() && (MyPgXact->vacuumFlags & PROC_IN_VACUUM))
 			xid = (TransactionId) BeginTranAutovacuumGTM();
 		else
 			xid = (TransactionId) BeginTranGTM(timestamp);
@@ -132,12 +134,12 @@ GetNewTransactionId(bool isSubXact)
 
 #ifdef PGXC
 	/* Only remote Coordinator or a Datanode accessed directly by an application can get a GXID */
-	if (IS_PGXC_COORDINATOR && !IsConnFromCoord() || IsPGXCNodeXactDatanodeDirect())
+	if ((IS_PGXC_COORDINATOR && !IsConnFromCoord()) || IsPGXCNodeXactDatanodeDirect())
 	{
 		if (TransactionIdIsValid(xid))
 		{
 			/* Log some information about the new transaction ID obtained */
-			if (IsAutoVacuumWorkerProcess() && (MyProc->vacuumFlags & PROC_IN_VACUUM))
+			if (IsAutoVacuumWorkerProcess() && (MyPgXact->vacuumFlags & PROC_IN_VACUUM))
 				elog(DEBUG1, "Assigned new transaction ID from GTM for autovacuum = %d", xid);
 			else
 				elog(DEBUG1, "Assigned new transaction ID from GTM = %d", xid);
@@ -172,7 +174,7 @@ GetNewTransactionId(bool isSubXact)
 			 * from snapshots so use a special function for this purpose.
 			 * For a simple worker get transaction ID like a normal transaction would do.
 			 */
-			if (MyProc->vacuumFlags & PROC_IN_VACUUM)
+			if (MyPgXact->vacuumFlags & PROC_IN_VACUUM)
 				next_xid = (TransactionId) BeginTranAutovacuumGTM();
 			else
 				next_xid = (TransactionId) BeginTranGTM(timestamp);
@@ -227,6 +229,28 @@ GetNewTransactionId(bool isSubXact)
 	 * to the DBA who somehow got past the earlier defenses).
 	 *----------
 	 */
+#ifdef PGXC
+	/*
+	 * In PG, the xid will never cross the wrap-around limit thanks to the
+	 * above checks. But in PGXC, if a new node is initialized and brought up,
+	 * it's own xid may not be in sync with the GTM gxid. The wrap-around limits
+	 * are initially set w.r.t. the last xid used. So if the Gxid-xid difference
+	 * is already more than 2^31, then the gxid is deemed to have already
+	 * crossed the wrap-around limit. So again in such cases as well, we should
+	 * allow only a standalone backend to run vacuum.
+	 */
+	if (TransactionIdFollowsOrEquals(xid, ShmemVariableCache->xidWrapLimit))
+	{
+		if (IsPostmasterEnvironment)
+		    ereport(ERROR,
+		       (errcode(ERRCODE_PROGRAM_LIMIT_EXCEEDED),
+		       errmsg("Xid wraparound might have already happened. database is not accepting commands on database with OID %u",
+		       ShmemVariableCache->oldestXidDB),
+		       errhint("Stop the postmaster and use a standalone backend to vacuum that database.\n"
+		               "You might also need to commit or roll back old prepared transactions.")));
+	}
+	else
+#endif
 	if (TransactionIdFollowsOrEquals(xid, ShmemVariableCache->xidVacLimit))
 	{
 		/*
@@ -254,27 +278,41 @@ GetNewTransactionId(bool isSubXact)
 		if (IsUnderPostmaster &&
 			TransactionIdFollowsOrEquals(xid, xidStopLimit))
 		{
+#ifdef PGXC
+			/*
+			 * Allow auto-vacuum to carry-on, so that it gets a chance to correct
+			 * the xid-wrap-limits w.r.t to gxid fetched from GTM.
+			 */
+			if (!IsAutoVacuumLauncherProcess() && !IsAutoVacuumWorkerProcess())
+			{
+				char  *oldest_datname = (OidIsValid(MyDatabaseId) ?
+			           get_database_name(oldest_datoid): NULL);
+#else
 			char	   *oldest_datname = get_database_name(oldest_datoid);
-
-			/* complain even if that DB has disappeared */
-			if (oldest_datname)
-				ereport(ERROR,
+#endif
+				/* complain even if that DB has disappeared */
+				if (oldest_datname)
+					ereport(ERROR,
 						(errcode(ERRCODE_PROGRAM_LIMIT_EXCEEDED),
 						 errmsg("database is not accepting commands to avoid wraparound data loss in database \"%s\"",
 								oldest_datname),
 						 errhint("Stop the postmaster and use a standalone backend to vacuum that database.\n"
 								 "You might also need to commit or roll back old prepared transactions.")));
-			else
-				ereport(ERROR,
+				else
+					ereport(ERROR,
 						(errcode(ERRCODE_PROGRAM_LIMIT_EXCEEDED),
 						 errmsg("database is not accepting commands to avoid wraparound data loss in database with OID %u",
 								oldest_datoid),
 						 errhint("Stop the postmaster and use a standalone backend to vacuum that database.\n"
 								 "You might also need to commit or roll back old prepared transactions.")));
+#ifdef PGXC
+			}
+#endif
 		}
 		else if (TransactionIdFollowsOrEquals(xid, xidWarnLimit))
 		{
-			char	   *oldest_datname = get_database_name(oldest_datoid);
+			char  *oldest_datname = (OidIsValid(MyDatabaseId) ?
+			           get_database_name(oldest_datoid): NULL);
 
 			/* complain even if that DB has disappeared */
 			if (oldest_datname)
@@ -324,13 +362,21 @@ GetNewTransactionId(bool isSubXact)
 	 * want the next incoming transaction to try it again.	We cannot assign
 	 * more XIDs until there is CLOG space for them.
 	 */
-#ifdef PGXC  /* defined(PGXC_COORD) || defined(PGXC_DATANODE) */
-	/* We may not be at the max, which is ok. Do not bother to increment.
-	 * We get this externally anyway, so it should not be needed in theory...
+#ifdef PGXC
+	/*
+	 * But first bring nextXid in sync with global xid. Actually we get xid
+	 * externally anyway, so it should not be needed to update nextXid in
+	 * theory, but it is required to keep nextXid close to the gxid
+	 * especially when vacuumfreeze is run using a standalone backend.
 	 */
-	if (increment_xid)
-#endif
+	if (increment_xid || !IsPostmasterEnvironment)
+	{
+		ShmemVariableCache->nextXid = xid;
 		TransactionIdAdvance(ShmemVariableCache->nextXid);
+	}
+#else
+	TransactionIdAdvance(ShmemVariableCache->nextXid);
+#endif
 
 	/*
 	 * We must store the new XID into the shared ProcArray before releasing
@@ -338,19 +384,19 @@ GetNewTransactionId(bool isSubXact)
 	 * latestCompletedXid is present in the ProcArray, which is essential for
 	 * correct OldestXmin tracking; see src/backend/access/transam/README.
 	 *
-	 * XXX by storing xid into MyProc without acquiring ProcArrayLock, we are
-	 * relying on fetch/store of an xid to be atomic, else other backends
+	 * XXX by storing xid into MyPgXact without acquiring ProcArrayLock, we
+	 * are relying on fetch/store of an xid to be atomic, else other backends
 	 * might see a partially-set xid here.	But holding both locks at once
 	 * would be a nasty concurrency hit.  So for now, assume atomicity.
 	 *
-	 * Note that readers of PGPROC xid fields should be careful to fetch the
+	 * Note that readers of PGXACT xid fields should be careful to fetch the
 	 * value only once, rather than assume they can read a value multiple
 	 * times and get the same answer each time.
 	 *
 	 * The same comments apply to the subxact xid count and overflow fields.
 	 *
-	 * A solution to the atomic-store problem would be to give each PGPROC its
-	 * own spinlock used only for fetching/storing that PGPROC's xid and
+	 * A solution to the atomic-store problem would be to give each PGXACT its
+	 * own spinlock used only for fetching/storing that PGXACT's xid and
 	 * related fields.
 	 *
 	 * If there's no room to fit a subtransaction XID into PGPROC, set the
@@ -372,20 +418,21 @@ GetNewTransactionId(bool isSubXact)
 		 * TransactionId and int fetch/store are atomic.
 		 */
 		volatile PGPROC *myproc = MyProc;
+		volatile PGXACT *mypgxact = MyPgXact;
 
 		if (!isSubXact)
-			myproc->xid = xid;
+			mypgxact->xid = xid;
 		else
 		{
-			int			nxids = myproc->subxids.nxids;
+			int			nxids = mypgxact->nxids;
 
 			if (nxids < PGPROC_MAX_CACHED_SUBXIDS)
 			{
 				myproc->subxids.xids[nxids] = xid;
-				myproc->subxids.nxids = nxids + 1;
+				mypgxact->nxids = nxids + 1;
 			}
 			else
-				myproc->subxids.overflowed = true;
+				mypgxact->overflowed = true;
 		}
 	}
 
