@@ -8,7 +8,7 @@
  * stepping on each others' toes.  Formerly we used table-level locks
  * on pg_database, but that's too coarse-grained.
  *
- * Portions Copyright (c) 1996-2011, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2012, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
@@ -26,7 +26,6 @@
 
 #include "access/genam.h"
 #include "access/heapam.h"
-#include "access/transam.h"
 #include "access/xact.h"
 #include "access/xlogutils.h"
 #include "catalog/catalog.h"
@@ -39,28 +38,27 @@
 #include "catalog/pg_tablespace.h"
 #include "commands/comment.h"
 #include "commands/dbcommands.h"
+#include "commands/seclabel.h"
 #include "commands/tablespace.h"
 #include "mb/pg_wchar.h"
 #include "miscadmin.h"
 #include "pgstat.h"
 #include "postmaster/bgwriter.h"
-#include "storage/bufmgr.h"
 #include "storage/copydir.h"
-#include "storage/fd.h"
 #include "storage/lmgr.h"
 #include "storage/ipc.h"
 #include "storage/procarray.h"
 #include "storage/smgr.h"
-#include "storage/standby.h"
 #include "utils/acl.h"
 #include "utils/builtins.h"
 #include "utils/fmgroids.h"
-#include "utils/lsyscache.h"
+#include "utils/memutils.h"
 #include "utils/pg_locale.h"
 #include "utils/snapmgr.h"
 #include "utils/syscache.h"
 #include "utils/tqual.h"
 #ifdef PGXC
+#include "pgxc/execRemote.h"
 #include "pgxc/pgxc.h"
 #include "access/gtm.h"
 #endif
@@ -75,6 +73,9 @@ typedef struct
 {
 	Oid			dest_dboid;		/* DB we are trying to move */
 	Oid			dest_tsoid;		/* tablespace we are trying to move to */
+#ifdef PGXC
+	Oid			src_tsoid;		/* tablespace we are trying to move from */
+#endif
 } movedb_failure_params;
 
 /* non-export function prototypes */
@@ -90,6 +91,9 @@ static bool have_createdb_privilege(void);
 static void remove_dbtablespaces(Oid db_id);
 static bool check_db_file_conflict(Oid db_id);
 static int	errdetail_busy_db(int notherbackends, int npreparedxacts);
+static void createdb_xact_callback(bool isCommit, void *arg);
+static void movedb_xact_callback(bool isCommit, void *arg);
+static void movedb_success_callback(Oid db_id, Oid tblspcoid);
 
 
 /*
@@ -130,6 +134,7 @@ createdb(const CreatedbStmt *stmt)
 	const char *dbtemplate = NULL;
 	char	   *dbcollate = NULL;
 	char	   *dbctype = NULL;
+	char	   *canonname;
 	int			encoding = -1;
 	int			dbconnlimit = -1;
 	int			notherbackends;
@@ -326,15 +331,17 @@ createdb(const CreatedbStmt *stmt)
 				(errcode(ERRCODE_WRONG_OBJECT_TYPE),
 				 errmsg("invalid server encoding %d", encoding)));
 
-	/* Check that the chosen locales are valid */
-	if (!check_locale(LC_COLLATE, dbcollate))
+	/* Check that the chosen locales are valid, and get canonical spellings */
+	if (!check_locale(LC_COLLATE, dbcollate, &canonname))
 		ereport(ERROR,
 				(errcode(ERRCODE_WRONG_OBJECT_TYPE),
-				 errmsg("invalid locale name %s", dbcollate)));
-	if (!check_locale(LC_CTYPE, dbctype))
+				 errmsg("invalid locale name: \"%s\"", dbcollate)));
+	dbcollate = canonname;
+	if (!check_locale(LC_CTYPE, dbctype, &canonname))
 		ereport(ERROR,
 				(errcode(ERRCODE_WRONG_OBJECT_TYPE),
-				 errmsg("invalid locale name %s", dbctype)));
+				 errmsg("invalid locale name: \"%s\"", dbctype)));
+	dbctype = canonname;
 
 	check_encoding_locale_matches(encoding, dbcollate, dbctype);
 
@@ -523,7 +530,8 @@ createdb(const CreatedbStmt *stmt)
 	copyTemplateDependencies(src_dboid, dboid);
 
 	/* Post creation hook for new database */
-	InvokeObjectAccessHook(OAT_POST_CREATE, DatabaseRelationId, dboid, 0);
+	InvokeObjectAccessHook(OAT_POST_CREATE,
+						   DatabaseRelationId, dboid, 0, NULL);
 
 	/*
 	 * Force a checkpoint before starting the copy. This will force dirty
@@ -680,9 +688,20 @@ createdb(const CreatedbStmt *stmt)
 	}
 	PG_END_ENSURE_ERROR_CLEANUP(createdb_failure_callback,
 								PointerGetDatum(&fparms));
-
 	/* Free our snapshot */
 	UnregisterSnapshot(snapshot);
+#ifdef PGXC
+	/*
+	 * Even if we are successful, ultimately this transaction can be aborted
+	 * because some other node failed. So arrange for cleanup on transaction
+	 * abort.
+	 * Unregistering snapshot above, will not have any problem when the
+	 * callback function is executed because the callback register's its
+	 * own snapshot in function remove_dbtablespaces
+	 */
+	set_dbcleanup_callback(createdb_xact_callback, &fparms.dest_dboid,
+	                      sizeof(fparms.dest_dboid));
+#endif
 }
 
 /*
@@ -722,11 +741,11 @@ check_encoding_locale_matches(int encoding, const char *collate, const char *cty
 		  (encoding == PG_SQL_ASCII && superuser())))
 		ereport(ERROR,
 				(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
-				 errmsg("encoding %s does not match locale %s",
+				 errmsg("encoding \"%s\" does not match locale \"%s\"",
 						pg_encoding_to_char(encoding),
 						ctype),
-			   errdetail("The chosen LC_CTYPE setting requires encoding %s.",
-						 pg_encoding_to_char(ctype_encoding))));
+		   errdetail("The chosen LC_CTYPE setting requires encoding \"%s\".",
+					 pg_encoding_to_char(ctype_encoding))));
 
 	if (!(collate_encoding == encoding ||
 		  collate_encoding == PG_SQL_ASCII ||
@@ -737,13 +756,32 @@ check_encoding_locale_matches(int encoding, const char *collate, const char *cty
 		  (encoding == PG_SQL_ASCII && superuser())))
 		ereport(ERROR,
 				(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
-				 errmsg("encoding %s does not match locale %s",
+				 errmsg("encoding \"%s\" does not match locale \"%s\"",
 						pg_encoding_to_char(encoding),
 						collate),
-			 errdetail("The chosen LC_COLLATE setting requires encoding %s.",
-					   pg_encoding_to_char(collate_encoding))));
+		 errdetail("The chosen LC_COLLATE setting requires encoding \"%s\".",
+				   pg_encoding_to_char(collate_encoding))));
 }
 
+#ifdef PGXC
+/*
+ * Error cleanup callback for createdb. Aftec createdb() succeeds, the
+ * transaction can still be aborted due to other nodes. So on abort-transaction,
+ * this function is called to do the cleanup. This involves removing directories
+ * created after successful completion.
+ * Nothing to be done on commit.
+ */
+static void
+createdb_xact_callback(bool isCommit, void *arg)
+{
+	if (isCommit)
+		return;
+
+	/* Throw away any successfully copied subdirectories */
+	remove_dbtablespaces(*(Oid *) arg);
+}
+#endif
+
 /* Error cleanup callback for createdb */
 static void
 createdb_failure_callback(int code, Datum arg)
@@ -811,6 +849,16 @@ dropdb(const char *dbname, bool missing_ok)
 		aclcheck_error(ACLCHECK_NOT_OWNER, ACL_KIND_DATABASE,
 					   dbname);
 
+	/* DROP hook for the database being removed */
+	if (object_access_hook)
+	{
+		ObjectAccessDrop drop_arg;
+
+		memset(&drop_arg, 0, sizeof(ObjectAccessDrop));
+		InvokeObjectAccessHook(OAT_DROP,
+							   DatabaseRelationId, db_id, 0, &drop_arg);
+	}
+
 	/*
 	 * Disallow dropping a DB that is marked istemplate.  This is just to
 	 * prevent people from accidentally dropping template0 or template1; they
@@ -852,9 +900,10 @@ dropdb(const char *dbname, bool missing_ok)
 	ReleaseSysCache(tup);
 
 	/*
-	 * Delete any comments associated with the database.
+	 * Delete any comments or security labels associated with the database.
 	 */
 	DeleteSharedComments(db_id, DatabaseRelationId);
+	DeleteSharedSecurityLabel(db_id, DatabaseRelationId);
 
 	/*
 	 * Remove settings associated with this database
@@ -879,18 +928,18 @@ dropdb(const char *dbname, bool missing_ok)
 	pgstat_drop_database(db_id);
 
 	/*
-	 * Tell bgwriter to forget any pending fsync and unlink requests for files
-	 * in the database; else the fsyncs will fail at next checkpoint, or
+	 * Tell checkpointer to forget any pending fsync and unlink requests for
+	 * files in the database; else the fsyncs will fail at next checkpoint, or
 	 * worse, it will delete files that belong to a newly created database
 	 * with the same OID.
 	 */
 	ForgetDatabaseFsyncRequests(db_id);
 
 	/*
-	 * Force a checkpoint to make sure the bgwriter has received the message
-	 * sent by ForgetDatabaseFsyncRequests. On Windows, this also ensures that
-	 * the bgwriter doesn't hold any open files, which would cause rmdir() to
-	 * fail.
+	 * Force a checkpoint to make sure the checkpointer has received the
+	 * message sent by ForgetDatabaseFsyncRequests. On Windows, this also
+	 * ensures that background procs don't hold any open files, which would
+	 * cause rmdir() to fail.
 	 */
 	RequestCheckpoint(CHECKPOINT_IMMEDIATE | CHECKPOINT_FORCE | CHECKPOINT_WAIT);
 
@@ -911,7 +960,7 @@ dropdb(const char *dbname, bool missing_ok)
 	 * according to pg_database, which is not good.
 	 */
 	ForceSyncCommit();
-
+	
 #ifdef PGXC
 	/* Drop sequences on gtm that are on the database dropped. */
 	if (IS_PGXC_COORDINATOR && !IsConnFromCoord())
@@ -1003,6 +1052,7 @@ RenameDatabase(const char *oldname, const char *newname)
 	heap_close(rel, NoLock);
 }
 
+
 #ifdef PGXC
 /*
  * IsSetTableSpace:
@@ -1015,12 +1065,14 @@ IsSetTableSpace(AlterDatabaseStmt *stmt)
 	/* Handle the SET TABLESPACE option separately */
 	foreach(option, stmt->options)
 	{
-		DefElem	*defel = (DefElem *) lfirst(option);
+		DefElem    *defel = (DefElem *) lfirst(option);
 		if (strcmp(defel->defname, "tablespace") == 0)
 			return true;
 	}
 	return false;
 }
+
+
 #endif
 
 /*
@@ -1072,6 +1124,15 @@ movedb(const char *dbname, const char *tblspcname)
 	LockSharedObjectForSession(DatabaseRelationId, db_id, 0,
 							   AccessExclusiveLock);
 
+#ifdef PGXC
+	/*
+	 * Now that we have session lock, transaction lock is not necessary.
+	 * Besides, a PREPARE does not allow both transaction and session lock on the
+	 * same object
+	 */
+	UnlockSharedObject(DatabaseRelationId, db_id, 0, AccessExclusiveLock);
+#endif
+
 	/*
 	 * Permission checks
 	 */
@@ -1146,7 +1207,7 @@ movedb(const char *dbname, const char *tblspcname)
 	 * process any pending unlink requests. Otherwise, the check for existing
 	 * files in the target directory might fail unnecessarily, not to mention
 	 * that the copy might fail due to source files getting deleted under it.
-	 * On Windows, this also ensures that the bgwriter doesn't hold any open
+	 * On Windows, this also ensures that background procs don't hold any open
 	 * files, which would cause rmdir() to fail.
 	 */
 	RequestCheckpoint(CHECKPOINT_IMMEDIATE | CHECKPOINT_FORCE | CHECKPOINT_WAIT);
@@ -1276,6 +1337,28 @@ movedb(const char *dbname, const char *tblspcname)
 	}
 	PG_END_ENSURE_ERROR_CLEANUP(movedb_failure_callback,
 								PointerGetDatum(&fparms));
+#ifdef PGXC
+	/*
+	 * Even if we are successful, ultimately this transaction may or may not
+	 * be committed. so arrange for cleanup of source directory or target
+	 * directory during commit or abort, respectively.
+	 */
+	fparms.src_tsoid = src_tblspcoid;
+	set_dbcleanup_callback(movedb_xact_callback, &fparms, sizeof(fparms));
+}
+
+/*
+ * movedb_success_callback:
+ * Cleanup files in the dbpath directory corresponding to db_id and tblspcoid.
+ * This function code is actual part of the movedb() operation in PG. We have
+ * made a function out of it for PGXC, and it gets called as part of the
+ * at-commit xact callback mechanism.
+ */
+static void
+movedb_success_callback(Oid db_id, Oid src_tblspcoid)
+{
+	char	   *src_dbpath = GetDatabasePath(db_id, src_tblspcoid);
+#endif /* PGXC */
 
 	/*
 	 * Commit the transaction so that the pg_database update is committed. If
@@ -1288,11 +1371,18 @@ movedb(const char *dbname, const char *tblspcname)
 	 * convinced it's a good idea; consider elog just after the transaction
 	 * really commits.
 	 */
+#ifdef PGXC
+	/*
+	 * Don't commit the transaction. We don't require the two separate
+	 * commits since we handle this function as at-commit xact callback.
+	 */
+#else
 	PopActiveSnapshot();
 	CommitTransactionCommand();
 
 	/* Start new transaction for the remaining work; don't need a snapshot */
 	StartTransactionCommand();
+#endif
 
 	/*
 	 * Remove files from the old tablespace
@@ -1323,8 +1413,34 @@ movedb(const char *dbname, const char *tblspcname)
 	/* Now it's safe to release the database lock */
 	UnlockSharedObjectForSession(DatabaseRelationId, db_id, 0,
 								 AccessExclusiveLock);
+
+#ifdef PGXC
+	pfree(src_dbpath);
+#endif
 }
 
+#ifdef PGXC
+/*
+ * Error cleanup callback for movedb. Aftec movedb() succeeds, the
+ * transaction can still be aborted due to other nodes. So on abort-transaction,
+ * this function is called to do the cleanup of target tablespace directory,
+ * and on transaction commit, it is called to cleanup source directory.
+ */
+static void
+movedb_xact_callback(bool isCommit, void *arg)
+{
+	movedb_failure_params *fparms = (movedb_failure_params *) DatumGetPointer(arg);
+
+	if (isCommit)
+		movedb_success_callback(fparms->dest_dboid, fparms->src_tsoid);
+	else
+	{
+		/* Call the same function that is used in ENSURE block for movedb() */
+		movedb_failure_callback(XACT_EVENT_ABORT, PointerGetDatum(arg));
+	}
+}
+#endif
+
 /* Error cleanup callback for movedb */
 static void
 movedb_failure_callback(int code, Datum arg)
@@ -1389,7 +1505,12 @@ AlterDatabase(AlterDatabaseStmt *stmt, bool isTopLevel)
 		/* currently, can't be specified along with any other options */
 		Assert(!dconnlimit);
 		/* this case isn't allowed within a transaction block */
-		PreventTransactionChain(isTopLevel, "ALTER DATABASE SET TABLESPACE");
+#ifdef PGXC
+		/* ... but we allow it on remote nodes */
+		if (IS_PGXC_COORDINATOR && !IsConnFromCoord())
+#endif
+			PreventTransactionChain(isTopLevel, "ALTER DATABASE SET TABLESPACE");
+
 		movedb(stmt->dbname, strVal(dtablespace->arg));
 		return;
 	}
@@ -1873,20 +1994,21 @@ check_db_file_conflict(Oid db_id)
 static int
 errdetail_busy_db(int notherbackends, int npreparedxacts)
 {
-	/*
-	 * We don't worry about singular versus plural here, since the English
-	 * rules for that don't translate very well.  But we can at least avoid
-	 * the case of zero items.
-	 */
 	if (notherbackends > 0 && npreparedxacts > 0)
+		/* We don't deal with singular versus plural here, since gettext
+		 * doesn't support multiple plurals in one string. */
 		errdetail("There are %d other session(s) and %d prepared transaction(s) using the database.",
 				  notherbackends, npreparedxacts);
 	else if (notherbackends > 0)
-		errdetail("There are %d other session(s) using the database.",
-				  notherbackends);
+		errdetail_plural("There is %d other session using the database.",
+						 "There are %d other sessions using the database.",
+						 notherbackends,
+						 notherbackends);
 	else
-		errdetail("There are %d prepared transaction(s) using the database.",
-				  npreparedxacts);
+		errdetail_plural("There is %d prepared transaction using the database.",
+						 "There are %d prepared transactions using the database.",
+						 npreparedxacts,
+						 npreparedxacts);
 	return 0;					/* just to keep ereport macro happy */
 }
 
