@@ -3,9 +3,9 @@
  * sequence.c
  *	  PostgreSQL sequences support code.
  *
- * Portions Copyright (c) 1996-2011, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2012, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
- * Portions Copyright (c) 2010-2013, Postgres-XC Development Group
+ * Portions Copyright (c) 2010-2012 Postgres-XC Development Group
  *
  *
  * IDENTIFICATION
@@ -15,9 +15,7 @@
  */
 #include "postgres.h"
 
-#include "access/heapam.h"
 #include "access/transam.h"
-#include "access/xact.h"
 #include "access/xlogutils.h"
 #include "catalog/dependency.h"
 #include "catalog/namespace.h"
@@ -28,7 +26,6 @@
 #include "funcapi.h"
 #include "miscadmin.h"
 #include "nodes/makefuncs.h"
-#include "storage/bufmgr.h"
 #include "storage/lmgr.h"
 #include "storage/proc.h"
 #include "storage/smgr.h"
@@ -515,8 +512,16 @@ AlterSequence(AlterSeqStmt *stmt)
 	bool			is_restart;
 #endif
 
-	/* open and AccessShareLock sequence */
-	relid = RangeVarGetRelid(stmt->sequence, false);
+	/* Open and lock sequence. */
+	relid = RangeVarGetRelid(stmt->sequence, AccessShareLock, stmt->missing_ok);
+	if (relid == InvalidOid)
+	{
+		ereport(NOTICE,
+				(errmsg("relation \"%s\" does not exist, skipping",
+						stmt->sequence->relname)));
+		return;
+	}
+
 	init_sequence(relid, &elm, &seqrel);
 
 	/* allow ALTER to sequence owner only */
@@ -533,6 +538,13 @@ AlterSequence(AlterSeqStmt *stmt)
 	/* Check and set new values */
 #ifdef PGXC
 	init_params(stmt->options, false, &new, &owned_by, &is_restart);
+
+	increment = new.increment_by;
+	min_value = new.min_value;
+	max_value = new.max_value;
+	start_value = new.start_value;
+	last_value = new.last_value;
+	cycle = new.is_cycled;
 #else
 	init_params(stmt->options, false, &new, &owned_by);
 #endif
@@ -546,15 +558,6 @@ AlterSequence(AlterSeqStmt *stmt)
 
 	memcpy(seq, &new, sizeof(FormData_pg_sequence));
 
-#ifdef PGXC
-	increment = new.increment_by;
-	min_value = new.min_value;
-	max_value = new.max_value;
-	start_value = new.start_value;
-	last_value = new.last_value;
-	cycle = new.is_cycled;
-#endif
-
 	MarkBufferDirty(buf);
 
 	/* XLOG stuff */
@@ -634,7 +637,16 @@ nextval(PG_FUNCTION_ARGS)
 	Oid			relid;
 
 	sequence = makeRangeVarFromNameList(textToQualifiedNameList(seqin));
-	relid = RangeVarGetRelid(sequence, false);
+
+	/*
+	 * XXX: This is not safe in the presence of concurrent DDL, but acquiring
+	 * a lock here is more expensive than letting nextval_internal do it,
+	 * since the latter maintains a cache that keeps us from hitting the lock
+	 * manager more than once per transaction.	It's not clear whether the
+	 * performance penalty is material in practice, but for now, we do it this
+	 * way.
+	 */
+	relid = RangeVarGetRelid(sequence, NoLock, false);
 
 	PG_RETURN_INT64(nextval_internal(relid));
 }
@@ -713,10 +725,6 @@ nextval_internal(Oid relid)
 		 * concurrency
 		 */
 		result = (int64) GetNextValGTM(seqname);
-		if (result < 0)
-			ereport(ERROR,
-					(errcode(ERRCODE_CONNECTION_FAILURE),
-					 errmsg("GTM error, could not obtain sequence value")));
 		pfree(seqname);
 
 		/* Update the on-disk data */
@@ -1886,9 +1894,9 @@ seq_redo(XLogRecPtr lsn, XLogRecord *record)
 	 * is also used for updating sequences, it's possible that a hot-standby
 	 * backend is examining the page concurrently; so we mustn't transiently
 	 * trash the buffer.  The solution is to build the correct new page
-	 * contents in local workspace and then memcpy into the buffer.  Then
-	 * only bytes that are supposed to change will change, even transiently.
-	 * We must palloc the local page for alignment reasons.
+	 * contents in local workspace and then memcpy into the buffer.  Then only
+	 * bytes that are supposed to change will change, even transiently. We
+	 * must palloc the local page for alignment reasons.
 	 */
 	localpage = (Page) palloc(BufferGetPageSize(buffer));
 
