@@ -3,7 +3,7 @@
  * nodeModifyTable.c
  *	  routines to handle ModifyTable nodes.
  *
- * Portions Copyright (c) 1996-2011, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2012, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
@@ -38,20 +38,35 @@
 #include "postgres.h"
 
 #include "access/xact.h"
+#ifdef PGXC
+#include "access/sysattr.h"
+#endif
 #include "commands/trigger.h"
 #include "executor/executor.h"
 #include "executor/nodeModifyTable.h"
 #include "miscadmin.h"
 #include "nodes/nodeFuncs.h"
 #ifdef PGXC
+#include "parser/parsetree.h"
 #include "pgxc/execRemote.h"
 #include "pgxc/pgxc.h"
 #endif
 #include "storage/bufmgr.h"
 #include "utils/builtins.h"
 #include "utils/memutils.h"
+#include "utils/rel.h"
 #include "utils/tqual.h"
 
+#ifdef PGXC
+static TupleTableSlot *fill_slot_with_oldvals(TupleTableSlot *slot, HeapTupleHeader oldtuphd, Bitmapset *modifiedCols);
+
+/* Copied from trigger.c */
+#define GetModifiedColumns(relinfo, estate) \
+	(rt_fetch((relinfo)->ri_RangeTableIndex, (estate)->es_range_table)->modifiedCols)
+/* Copied from tid.c */
+#define DatumGetItemPointer(X)	 ((ItemPointer) DatumGetPointer(X))
+#endif
+
 
 /*
  * Verify that the tuples to be produced by INSERT or UPDATE match the
@@ -173,7 +188,7 @@ ExecInsert(TupleTableSlot *slot,
 	Oid			newId;
 	List	   *recheckIndexes = NIL;
 #ifdef PGXC	
-	PlanState  *resultRemoteRel = NULL;
+	RemoteQueryState  *resultRemoteRel = NULL;
 #endif
 
 	/*
@@ -188,7 +203,7 @@ ExecInsert(TupleTableSlot *slot,
 	resultRelInfo = estate->es_result_relation_info;
 	resultRelationDesc = resultRelInfo->ri_RelationDesc;
 #ifdef PGXC	
-	resultRemoteRel = estate->es_result_remoterel;
+	resultRemoteRel = (RemoteQueryState *) estate->es_result_remoterel;
 #endif
 	/*
 	 * If the result relation has OIDs, force the tuple's OID to zero so that
@@ -243,8 +258,7 @@ ExecInsert(TupleTableSlot *slot,
 #ifdef PGXC
 		if (IS_PGXC_COORDINATOR && resultRemoteRel)
 		{
-			ExecRemoteQueryStandard(resultRelationDesc, (RemoteQueryState *)resultRemoteRel, slot);
-
+			slot = ExecProcNodeDMLInXC(estate, planSlot, slot);
 			/*
 			 * PGXCTODO: If target table uses WITH OIDS, this should be set to the Oid inserted
 			 * but Oids are not consistent among nodes in Postgres-XC, so this is set to the
@@ -276,6 +290,11 @@ ExecInsert(TupleTableSlot *slot,
 
 	if (canSetTag)
 	{
+#ifdef PGXC
+		if (IS_PGXC_COORDINATOR && resultRemoteRel)
+			estate->es_processed += resultRemoteRel->rqs_processed;
+		else
+#endif
 		(estate->es_processed)++;
 		estate->es_lastoid = newId;
 		setLastTid(&(tuple->t_self));
@@ -288,8 +307,16 @@ ExecInsert(TupleTableSlot *slot,
 
 	/* Process RETURNING if present */
 	if (resultRelInfo->ri_projectReturning)
+#ifdef PGXC
+	{
+		if (TupIsNull(slot))
+			return NULL;
+#endif
 		return ExecProcessReturning(resultRelInfo->ri_projectReturning,
 									slot, planSlot);
+#ifdef PGXC
+	}
+#endif
 
 	return NULL;
 }
@@ -322,7 +349,8 @@ ExecDelete(ItemPointer tupleid,
 	ItemPointerData update_ctid;
 	TransactionId update_xmax;
 #ifdef PGXC
-	PlanState  *resultRemoteRel = NULL;
+	RemoteQueryState  *resultRemoteRel = NULL;
+	TupleTableSlot *slot = NULL;
 #endif
 
 	/*
@@ -331,7 +359,7 @@ ExecDelete(ItemPointer tupleid,
 	resultRelInfo = estate->es_result_relation_info;
 	resultRelationDesc = resultRelInfo->ri_RelationDesc;
 #ifdef PGXC
-	resultRemoteRel = estate->es_result_remoterel;
+	resultRemoteRel = (RemoteQueryState *) estate->es_result_remoterel;
 #endif
 
 	/* BEFORE ROW DELETE Triggers */
@@ -341,8 +369,10 @@ ExecDelete(ItemPointer tupleid,
 		bool		dodelete;
 
 		dodelete = ExecBRDeleteTriggers(estate, epqstate, resultRelInfo,
+#ifdef PGXC
+										oldtuple,
+#endif
 										tupleid);
-
 		if (!dodelete)			/* "do nothing" */
 			return NULL;
 	}
@@ -383,7 +413,7 @@ ldelete:;
 #ifdef PGXC
 		if (IS_PGXC_COORDINATOR && resultRemoteRel)
 		{
-			ExecRemoteQueryStandard(resultRelationDesc, (RemoteQueryState *)resultRemoteRel, planSlot);
+			slot = ExecProcNodeDMLInXC(estate, planSlot, NULL);
 		}
 		else
 		{
@@ -446,18 +476,37 @@ ldelete:;
 	}
 
 	if (canSetTag)
+#ifdef PGXC
+	{
+		if (IS_PGXC_COORDINATOR && resultRemoteRel)
+			estate->es_processed += resultRemoteRel->rqs_processed;
+		else
+#endif
 		(estate->es_processed)++;
+#ifdef PGXC
+	}
+#endif
 
 #ifdef PGXC
-	/*
-	 * Do not fire triggers on remote relation, it would not find old tuple
-	 */
-	if (resultRemoteRel == NULL)
-#endif
+	ExecARDeleteTriggers(estate, resultRelInfo, oldtuple, tupleid);
+#else
 	/* AFTER ROW DELETE Triggers */
 	ExecARDeleteTriggers(estate, resultRelInfo, tupleid);
+#endif
 
 	/* Process RETURNING if present */
+#ifdef PGXC
+	if (IS_PGXC_COORDINATOR && resultRemoteRel &&
+		resultRelInfo->ri_projectReturning)
+	{
+		if (TupIsNull(slot))
+			return NULL;
+
+		return ExecProcessReturning(resultRelInfo->ri_projectReturning,
+						slot, planSlot);
+	}
+	else
+#endif
 	if (resultRelInfo->ri_projectReturning)
 	{
 		/*
@@ -546,7 +595,7 @@ ExecUpdate(ItemPointer tupleid,
 	TransactionId update_xmax;
 	List	   *recheckIndexes = NIL;
 #ifdef PGXC
-	PlanState  *resultRemoteRel = NULL;
+	RemoteQueryState  *resultRemoteRel = NULL;
 #endif
 
 	/*
@@ -555,6 +604,22 @@ ExecUpdate(ItemPointer tupleid,
 	if (IsBootstrapProcessingMode())
 		elog(ERROR, "cannot UPDATE during bootstrap");
 
+#ifdef PGXC
+	resultRemoteRel = (RemoteQueryState *) estate->es_result_remoterel;
+
+	/*
+	 * For remote tables, the plan slot does not have all NEW tuple values in
+	 * the plan slot. If oldtuple is supplied, we would also need a complete
+	 * NEW tuple. Currently for remote tables, triggers are the only case where
+	 * oldtuple is passed. Craft the NEW tuple using OLD tuple and updated
+	 * values from NEW tuple slot, and store the NEW tuple back into the NEW
+	 * tuple slot.
+	 */
+	if (IS_PGXC_COORDINATOR && resultRemoteRel && oldtuple != NULL)
+		slot = fill_slot_with_oldvals(slot, oldtuple,
+				GetModifiedColumns(estate->es_result_relation_info, estate));
+#endif
+
 	/*
 	 * get the heap tuple out of the tuple table slot, making sure we have a
 	 * writable copy
@@ -566,16 +631,19 @@ ExecUpdate(ItemPointer tupleid,
 	 */
 	resultRelInfo = estate->es_result_relation_info;
 	resultRelationDesc = resultRelInfo->ri_RelationDesc;
-#ifdef PGXC
-	resultRemoteRel = estate->es_result_remoterel;
-#endif
 
 	/* BEFORE ROW UPDATE Triggers */
 	if (resultRelInfo->ri_TrigDesc &&
 		resultRelInfo->ri_TrigDesc->trig_update_before_row)
 	{
+#ifdef PGXC
+
+		slot = ExecBRUpdateTriggers(estate, epqstate, resultRelInfo,
+									oldtuple, tupleid, slot);
+#else
 		slot = ExecBRUpdateTriggers(estate, epqstate, resultRelInfo,
 									tupleid, slot);
+#endif
 
 		if (slot == NULL)		/* "do nothing" */
 			return NULL;
@@ -626,7 +694,7 @@ lreplace:;
 #ifdef PGXC
 		if (IS_PGXC_COORDINATOR && resultRemoteRel)
 		{
-			ExecRemoteQueryStandard(resultRelationDesc, (RemoteQueryState *)resultRemoteRel, planSlot);
+			slot = ExecProcNodeDMLInXC(estate, planSlot, slot);
 		}
 		else
 		{
@@ -710,24 +778,38 @@ lreplace:;
 	}
 
 	if (canSetTag)
-		(estate->es_processed)++;
-
 #ifdef PGXC
-	/*
-	 * Do not fire triggers on remote relation, it would not find old tuple
-	 */
-	if (resultRemoteRel == NULL)
+	{
+		if (IS_PGXC_COORDINATOR && resultRemoteRel)
+			estate->es_processed += resultRemoteRel->rqs_processed;
+		else
 #endif
+		(estate->es_processed)++;
+#ifdef PGXC
+	}
+#endif
+
 	/* AFTER ROW UPDATE Triggers */
 	ExecARUpdateTriggers(estate, resultRelInfo, tupleid, tuple,
+#ifdef PGXC
+						 oldtuple,
+#endif
 						 recheckIndexes);
 
 	list_free(recheckIndexes);
 
 	/* Process RETURNING if present */
 	if (resultRelInfo->ri_projectReturning)
+#ifdef PGXC
+	{
+		if (TupIsNull(slot))
+			return NULL;
+#endif
 		return ExecProcessReturning(resultRelInfo->ri_projectReturning,
 									slot, planSlot);
+#ifdef PGXC
+	}
+#endif
 
 	return NULL;
 }
@@ -927,6 +1009,21 @@ ExecModifyTable(ModifyTableState *node)
 					tuple_ctid = *tupleid;		/* be sure we don't free
 												 * ctid!! */
 					tupleid = &tuple_ctid;
+
+#ifdef PGXC
+					/* If available, also extract the OLD row */
+					if (IS_PGXC_COORDINATOR &&
+						RelationGetLocInfo(resultRelInfo->ri_RelationDesc) &&
+					    junkfilter->jf_xc_wholerow != InvalidAttrNumber)
+					{
+						datum = ExecGetJunkAttribute(slot,
+													 junkfilter->jf_xc_wholerow,
+													 &isNull);
+
+						if (!isNull)
+							oldtuple = DatumGetHeapTupleHeader(datum);
+					}
+#endif
 				}
 				else
 				{
@@ -947,7 +1044,9 @@ ExecModifyTable(ModifyTableState *node)
 			if (operation != CMD_DELETE)
 				slot = ExecFilterJunk(junkfilter, slot);
 		}
-
+#ifdef PGXC
+		estate->es_result_remoterel = remoterelstate;
+#endif
 		switch (operation)
 		{
 			case CMD_INSERT:
@@ -973,6 +1072,9 @@ ExecModifyTable(ModifyTableState *node)
 		if (slot)
 		{
 			estate->es_result_relation_info = saved_resultRelInfo;
+#ifdef PGXC
+			estate->es_result_remoterel = saved_resultRemoteRel;
+#endif
 			return slot;
 		}
 	}
@@ -1009,6 +1111,9 @@ ExecInitModifyTable(ModifyTable *node, EState *estate, int eflags)
 	Plan	   *subplan;
 	ListCell   *l;
 	int			i;
+#ifdef PGXC
+	PlanState  *saved_remoteRelInfo;
+#endif
 
 	/* check for unsupported flags */
 	Assert(!(eflags & (EXEC_FLAG_BACKWARD | EXEC_FLAG_MARK)));
@@ -1046,6 +1151,9 @@ ExecInitModifyTable(ModifyTable *node, EState *estate, int eflags)
 	 * sub-plan; ExecContextForcesOids depends on that!
 	 */
 	saved_resultRelInfo = estate->es_result_relation_info;
+#ifdef PGXC
+	saved_remoteRelInfo = estate->es_result_remoterel;
+#endif
 
 	resultRelInfo = mtstate->resultRelInfo;
 	i = 0;
@@ -1071,8 +1179,8 @@ ExecInitModifyTable(ModifyTable *node, EState *estate, int eflags)
 		 * If there are indices on the result relation, open them and save
 		 * descriptors in the result relation info, so that we can add new
 		 * index entries for the tuples we add/update.	We need not do this
-		 * for a DELETE, however, since deletion doesn't affect indexes.
-		 * Also, inside an EvalPlanQual operation, the indexes might be open
+		 * for a DELETE, however, since deletion doesn't affect indexes. Also,
+		 * inside an EvalPlanQual operation, the indexes might be open
 		 * already, since we share the resultrel state with the original
 		 * query.
 		 */
@@ -1101,6 +1209,9 @@ ExecInitModifyTable(ModifyTable *node, EState *estate, int eflags)
 	}
 
 	estate->es_result_relation_info = saved_resultRelInfo;
+#ifdef PGXC
+	estate->es_result_remoterel = saved_remoteRelInfo;
+#endif
 
 	/*
 	 * Initialize RETURNING projections if needed.
@@ -1255,6 +1366,20 @@ ExecInitModifyTable(ModifyTable *node, EState *estate, int eflags)
 						j->jf_junkAttNo = ExecFindJunkAttribute(j, "ctid");
 						if (!AttributeNumberIsValid(j->jf_junkAttNo))
 							elog(ERROR, "could not find junk ctid column");
+
+#ifdef PGXC
+						if (IS_PGXC_COORDINATOR &&
+							RelationGetLocInfo(resultRelInfo->ri_RelationDesc))
+						{
+							/*
+							 * We may or may not need these attributes depending upon
+							 * the exact kind of trigger. We defer the check; instead throw
+							 * error only at the point when we need but don't find one.
+							 */
+							j->jf_xc_node_id = ExecFindJunkAttribute(j, "xc_node_id");
+							j->jf_xc_wholerow = ExecFindJunkAttribute(j, "wholerow");
+						}
+#endif
 					}
 					else
 					{
@@ -1344,3 +1469,58 @@ ExecReScanModifyTable(ModifyTableState *node)
 	 */
 	elog(ERROR, "ExecReScanModifyTable is not implemented");
 }
+
+#ifdef PGXC
+
+/*
+ * fill_slot_with_oldvals:
+ * Create a new tuple using the existing 'oldtuphd' and new data from
+ * 'replace_slot'. So the values of the modified attributes are taken from
+ * replace_slot, and overwritten onto the oldtuphd. Finally the new tuple is
+ * stored in 'replace_slot'. This is a convenience function for generating
+ * the NEW tuple given the plan slot and old tuple.
+ */
+static TupleTableSlot *
+fill_slot_with_oldvals(TupleTableSlot *replace_slot, HeapTupleHeader oldtuphd, Bitmapset *modifiedCols)
+{
+	HeapTupleData oldtuple;
+	HeapTuple newtuple;
+	int natts = replace_slot->tts_tupleDescriptor->natts;
+	int attindex;
+	bool *replaces;
+
+	if (!oldtuphd)
+		elog(ERROR, "expected valid OLD tuple for triggers");
+		
+	oldtuple.t_data = oldtuphd;
+	oldtuple.t_len = HeapTupleHeaderGetDatumLength(oldtuphd);
+	ItemPointerSetInvalid(&(oldtuple.t_self));
+	oldtuple.t_tableOid = InvalidOid;
+	oldtuple.t_xc_node_id = 0;
+
+	replaces = (bool *) palloc0(natts * sizeof(bool));
+	for (attindex = 0; attindex < natts; attindex++)
+	{
+		if (bms_is_member(attindex + 1 - FirstLowInvalidHeapAttributeNumber,
+							  modifiedCols))
+			replaces[attindex] = true;
+		else
+			replaces[attindex] = false;
+	}
+
+	slot_getallattrs(replace_slot);
+
+	newtuple = heap_modify_tuple(&oldtuple, replace_slot->tts_tupleDescriptor,
+	                             replace_slot->tts_values,
+	                             replace_slot->tts_isnull,
+	                             replaces);
+
+	pfree(replaces);
+
+	/*
+	 * Ultimately store the tuple in the same slot from where we retrieved
+	 * values to be replaced.
+	 */
+	return ExecStoreTuple(newtuple, replace_slot, InvalidBuffer, false);
+}
+#endif
