@@ -7,7 +7,7 @@
  *
  *
  * Portions Copyright (c) 1996-2009, PostgreSQL Global Development Group
- * Portions Copyright (c) 2010-2013, Postgres-XC Development Group
+ * Portions Copyright (c) 2010-2012 Postgres-XC Development Group
  *
  * IDENTIFICATION
  *	  $$
@@ -28,6 +28,7 @@
 #include "access/gtm.h"
 #include "access/transam.h"
 #include "access/xact.h"
+#include "catalog/pg_type.h"
 #include "commands/prepare.h"
 #include "gtm/gtm_c.h"
 #include "nodes/nodes.h"
@@ -44,11 +45,13 @@
 #include "utils/elog.h"
 #include "utils/memutils.h"
 #include "utils/fmgroids.h"
+#include "utils/snapmgr.h"
 #include "utils/syscache.h"
 #include "utils/lsyscache.h"
 #include "utils/formatting.h"
 #include "../interfaces/libpq/libpq-fe.h"
 
+#define CMD_ID_MSG_LEN 8
 
 /* Number of connections held */
 static int	datanode_count = 0;
@@ -1129,7 +1132,11 @@ pgxc_node_send_parse(PGXCNodeHandle * handle, const char* statement,
 	/* message length */
 	int			msgLen;
 	int			cnt_params;
+#if 0
 	size_t		old_outEnd = handle->outEnd;
+#else
+#define old_outEnd handle->outEnd
+#endif	
 
 	/* if there are parameters, param_types should exist */
 	Assert(num_params <= 0 || param_types);
@@ -1138,7 +1145,15 @@ pgxc_node_send_parse(PGXCNodeHandle * handle, const char* statement,
 	/* find names of the types of parameters */
 	for (cnt_params = 0; cnt_params < num_params; cnt_params++)
 	{
-		paramTypes[cnt_params] = format_type_be(param_types[cnt_params]);
+		Oid typeoid;
+
+		/* Parameters with no types are simply ignored */
+		if (OidIsValid(param_types[cnt_params]))
+			typeoid = param_types[cnt_params];
+		else
+			typeoid = INT4OID;
+
+		paramTypes[cnt_params] = format_type_be(typeoid);
 		paramTypeLen += strlen(paramTypes[cnt_params]) + 1;
 	}
 
@@ -1596,6 +1611,40 @@ pgxc_node_send_gxid(PGXCNodeHandle *handle, GlobalTransactionId gxid)
 	return 0;
 }
 
+/*
+ * Send the Command ID down to the PGXC node
+ */
+int
+pgxc_node_send_cmd_id(PGXCNodeHandle *handle, CommandId cid)
+{
+	int			msglen = CMD_ID_MSG_LEN;
+	int			i32;
+
+	/* No need to send command ID if its sending flag is not enabled */
+	if (!IsSendCommandId())
+		return 0;
+
+	/* Invalid connection state, return error */
+	if (handle->state != DN_CONNECTION_STATE_IDLE)
+		return EOF;
+
+	/* msgType + msgLen */
+	if (ensure_out_buffer_capacity(handle->outEnd + 1 + msglen, handle) != 0)
+	{
+		add_error_message(handle, "out of memory");
+		return EOF;
+	}
+
+	handle->outBuffer[handle->outEnd++] = 'M';
+	msglen = htonl(msglen);
+	memcpy(handle->outBuffer + handle->outEnd, &msglen, 4);
+	handle->outEnd += 4;
+	i32 = htonl(cid);
+	memcpy(handle->outBuffer + handle->outEnd, &i32, 4);
+	handle->outEnd += 4;
+
+	return 0;
+}
 
 /*
  * Send the snapshot down to the PGXC node
@@ -1636,7 +1685,7 @@ pgxc_node_send_snapshot(PGXCNodeHandle *handle, Snapshot snapshot)
 	memcpy(handle->outBuffer + handle->outEnd, &nval, 4);
 	handle->outEnd += 4;
 
-	nval = htonl(snapshot->recent_global_xmin);
+	nval = htonl(RecentGlobalXmin);
 	memcpy(handle->outBuffer + handle->outEnd, &nval, 4);
 	handle->outEnd += 4;
 
