@@ -3,7 +3,7 @@
  * rewriteHandler.c
  *		Primary module of query rewriter.
  *
- * Portions Copyright (c) 1996-2011, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2012, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  * IDENTIFICATION
@@ -13,9 +13,9 @@
  */
 #include "postgres.h"
 
-#include "access/heapam.h"
 #include "access/sysattr.h"
 #include "catalog/pg_type.h"
+#include "commands/trigger.h"
 #include "nodes/makefuncs.h"
 #include "nodes/nodeFuncs.h"
 #include "parser/analyze.h"
@@ -26,13 +26,12 @@
 #include "rewrite/rewriteManip.h"
 #include "utils/builtins.h"
 #include "utils/lsyscache.h"
-#include "commands/trigger.h"
+#include "utils/rel.h"
 
 #ifdef PGXC
 #include "pgxc/locator.h"
 #include "pgxc/nodemgr.h"
 #include "pgxc/pgxc.h"
-#include "pgxc/postgresql_fdw.h"
 #include "nodes/nodes.h"
 #include "optimizer/planner.h"
 #include "optimizer/var.h"
@@ -1325,29 +1324,47 @@ rewriteTargetListUD(Query *parsetree, RangeTblEntry *target_rte,
 	parsetree->targetList = lappend(parsetree->targetList, tle);
 
 #ifdef PGXC
-	/*
-	 * If relation is non-replicated, we need also to identify the Datanode
-	 * from where tuple is fetched.
-	 */
-	if (IS_PGXC_COORDINATOR &&
-		!IsConnFromCoord() &&
-		!IsLocatorReplicated(GetRelationLocType(RelationGetRelid(target_relation))))
+	/* Add further attributes required for Coordinator */
+
+	if (IS_PGXC_COORDINATOR && RelationGetLocInfo(target_relation) != NULL
+		&& target_relation->rd_rel->relkind == RELKIND_RELATION)
 	{
-		var = makeVar(parsetree->resultRelation,
-					  XC_NodeIdAttributeNumber,
-					  INT4OID,
-					  -1,
-					  InvalidOid,
-					  0);
+		/*
+		 * If relation is non-replicated, we need also to identify the Datanode
+		 * from where tuple is fetched.
+		 */
+		if (!IsRelationReplicated(RelationGetLocInfo(target_relation)))
+		{
+			var = makeVar(parsetree->resultRelation,
+						  XC_NodeIdAttributeNumber,
+						  INT4OID,
+						  -1,
+						  InvalidOid,
+						  0);
 
-		attrname = "xc_node_id";
+			tle = makeTargetEntry((Expr *) var,
+								  list_length(parsetree->targetList) + 1,
+								  pstrdup("xc_node_id"),
+								  true);
 
-		tle = makeTargetEntry((Expr *) var,
-							  list_length(parsetree->targetList) + 1,
-							  pstrdup(attrname),
-							  true);
+			parsetree->targetList = lappend(parsetree->targetList, tle);
+		}
 
-		parsetree->targetList = lappend(parsetree->targetList, tle);		
+		/* For non-shippable triggers, we need OLD row. */
+		if (pgxc_trig_oldrow_reqd(target_relation,
+								  parsetree->commandType))
+		{
+			var = makeWholeRowVar(target_rte,
+								  parsetree->resultRelation,
+								  0,
+								  false);
+
+			tle = makeTargetEntry((Expr *) var,
+				  list_length(parsetree->targetList) + 1,
+				  pstrdup("wholerow"),
+				  true);
+			parsetree->targetList = lappend(parsetree->targetList, tle);
+		}
 	}
 #endif
 }
@@ -1532,6 +1549,7 @@ ApplyRetrieveRule(Query *parsetree,
 
 	rte->rtekind = RTE_SUBQUERY;
 	rte->relid = InvalidOid;
+	rte->security_barrier = RelationIsSecurityView(relation);
 	rte->subquery = rule_action;
 	rte->inh = false;			/* must not be set for a subquery */
 
@@ -2140,7 +2158,7 @@ RewriteQuery(Query *parsetree, List *rewrite_events)
 		}
 		else
 			elog(ERROR, "unrecognized commandType: %d", (int) event);
-		
+
 #ifdef PGXC
 		if (parsetree_list == NIL)
 		{
@@ -2255,7 +2273,7 @@ RewriteQuery(Query *parsetree, List *rewrite_events)
 			foreach(pt_cell, parsetree_list)
 			{
 				Query  *query;
-				
+
 				query = (Query *)lfirst(pt_cell);
 
 				/*
@@ -2276,7 +2294,7 @@ RewriteQuery(Query *parsetree, List *rewrite_events)
 										&instead,
 										&returning,
 										&qual_product);
-					
+
 					qual_product_list = lappend(qual_product_list,  qual_product);
 
 					/*
@@ -2399,14 +2417,14 @@ RewriteQuery(Query *parsetree, List *rewrite_events)
 	}
 	else
 	{
-		int query_no = 0;	
+		int query_no = 0;
 
 		foreach(pt_cell, parsetree_list)
 		{
 
 			Query	*query = NULL;
 			Query	*qual = NULL;
-			
+
 			query = (Query *)lfirst(pt_cell);
 			if (!instead)
 			{
@@ -2475,6 +2493,7 @@ RewriteQuery(Query *parsetree, List *rewrite_events)
 List *
 QueryRewrite(Query *parsetree)
 {
+	uint32		input_query_id = parsetree->queryId;
 	List	   *querylist;
 	List	   *results;
 	ListCell   *l;
@@ -2499,6 +2518,8 @@ QueryRewrite(Query *parsetree)
 	 * Step 2
 	 *
 	 * Apply all the RIR rules on each query
+	 *
+	 * This is also a handy place to mark each query with the original queryId
 	 */
 	results = NIL;
 	foreach(l, querylist)
@@ -2506,6 +2527,9 @@ QueryRewrite(Query *parsetree)
 		Query	   *query = (Query *) lfirst(l);
 
 		query = fireRIRrules(query, NIL, false);
+
+		query->queryId = input_query_id;
+
 		results = lappend(results, query);
 	}
 
@@ -2576,25 +2600,38 @@ QueryRewriteCTAS(Query *parsetree)
 	StringInfoData cquery;
 	ListCell *col;
 	Query *cparsetree;
-	List *raw_parsetree_list;
+	List *raw_parsetree_list, *tlist;
 	char *selectstr;
+	CreateTableAsStmt *stmt;
+	IntoClause *into;
+	ListCell *lc;
 
-	if (parsetree->commandType != CMD_SELECT ||
-		(parsetree->intoClause == NULL))
+	if (parsetree->commandType != CMD_UTILITY ||
+		!IsA(parsetree->utilityStmt, CreateTableAsStmt))
 		elog(ERROR, "Unexpected commandType or intoClause is not set properly");
 
 	/* Get the target table */
-	relation = parsetree->intoClause->rel;
+	stmt = (CreateTableAsStmt *) parsetree->utilityStmt;
+	relation = stmt->into->rel;
 
 	/* Start building a CreateStmt for creating the target table */
-   	create_stmt = makeNode(CreateStmt);
+	create_stmt = makeNode(CreateStmt);
 	create_stmt->relation = relation;
+	into = stmt->into;
 
-	/* 
+	/* Obtain the target list of new table */
+	Assert(IsA(stmt->query, Query));
+	cparsetree = (Query *) stmt->query;
+	tlist = cparsetree->targetList;
+
+	/*
 	 * Based on the targetList, populate the column information for the target
-	 * table.
+	 * table. If a column name list was specified in CREATE TABLE AS, override
+	 * the column names derived from the query. (Too few column names are OK, too
+	 * many are not.).
 	 */
-	foreach(col, parsetree->targetList)
+	lc = list_head(into->colNames);
+	foreach(col, tlist)
 	{
 		TargetEntry *tle = (TargetEntry *)lfirst(col);
 		ColumnDef   *coldef;
@@ -2607,6 +2644,15 @@ QueryRewriteCTAS(Query *parsetree)
 		coldef = makeNode(ColumnDef);
 		typename = makeNode(TypeName);
 
+		/* Take the column name specified if any */
+		if (lc)
+		{
+			coldef->colname = strVal(lfirst(lc));
+			lc = lnext(lc);
+		}
+		else
+			coldef->colname = pstrdup(tle->resname);
+
 		coldef->inhcount = 0;
 		coldef->is_local = true;
 		coldef->is_not_null = false;
@@ -2614,7 +2660,7 @@ QueryRewriteCTAS(Query *parsetree)
 		coldef->cooked_default = NULL;
 		coldef->constraints = NIL;
 
-		/* 
+		/*
 		 * Set typeOid and typemod. The name of the type is derived while
 		 * generating query
 		 */
@@ -2623,25 +2669,25 @@ QueryRewriteCTAS(Query *parsetree)
 
 		coldef->typeName = typename;
 
-		/* 
-		 * XXX Should we look at the column specifications in the intoClause
-		 * instead of the target entry ?
-		 */
-		coldef->colname = pstrdup(tle->resname);
 		tableElts = lappend(tableElts, coldef);
 	}
 
-	/* 
+	if (lc != NULL)
+		ereport(ERROR,
+				(errcode(ERRCODE_SYNTAX_ERROR),
+				 errmsg("CREATE TABLE AS specifies too many column names")));
+
+	/*
 	 * Set column information and the distribution mechanism (which will be
 	 * NULL for SELECT INTO and the default mechanism will be picked)
 	 */
 	create_stmt->tableElts = tableElts;
-	create_stmt->distributeby = parsetree->intoClause->distributeby;
-	create_stmt->subcluster = parsetree->intoClause->subcluster;
+	create_stmt->distributeby = stmt->into->distributeby;
+	create_stmt->subcluster = stmt->into->subcluster;
 
-	create_stmt->tablespacename = parsetree->intoClause->tableSpaceName;
-	create_stmt->oncommit = parsetree->intoClause->onCommit;
-	create_stmt->options = parsetree->intoClause->options;
+	create_stmt->tablespacename = stmt->into->tableSpaceName;
+	create_stmt->oncommit = stmt->into->onCommit;
+	create_stmt->options = stmt->into->options;
 
 	/*
 	 * Check consistency of arguments
@@ -2655,7 +2701,7 @@ QueryRewriteCTAS(Query *parsetree)
 	/* Get a copy of the parsetree which we can freely modify  */
 	cparsetree = copyObject(parsetree);
 
-	/* 
+	/*
 	 * Now build a utility statement in order to run the CREATE TABLE DDL on
 	 * the local and remote nodes. We keep others fields as it is since they
 	 * are ignored anyways by deparse_query.
@@ -2664,7 +2710,7 @@ QueryRewriteCTAS(Query *parsetree)
 	cparsetree->utilityStmt = (Node *) create_stmt;
 
 	initStringInfo(&cquery);
-	deparse_query(cparsetree, &cquery, NIL);
+	deparse_query(cparsetree, &cquery, NIL, false, false);
 
 	/* Finally, fire off the query to run the DDL */
 	ProcessUtility(cparsetree->utilityStmt, cquery.data, NULL, true, NULL,
@@ -2672,14 +2718,14 @@ QueryRewriteCTAS(Query *parsetree)
 					NULL);
 
 	/*
-	 * Now fold the SELECT statement into an INSERT INTO statement. The
-	 * intoClause is no more required.
+	 * Now fold the CTAS statement into an INSERT INTO statement. The
+	 * utility is no more required.
 	 */
-	parsetree->intoClause = NULL;
+	parsetree->utilityStmt = NULL;
 
 	/* Get the SELECT query string */
 	initStringInfo(&cquery);
-	deparse_query(parsetree, &cquery, NIL);
+	deparse_query((Query *)stmt->query, &cquery, NIL, true, false);
 	selectstr = pstrdup(cquery.data);
 
 	/* Now, finally build the INSERT INTO statement */
@@ -2699,3 +2745,4 @@ QueryRewriteCTAS(Query *parsetree)
 }
 #endif
 
+
