@@ -3,7 +3,7 @@
  * lockfuncs.c
  *		Functions for SQL access to various lock-manager capabilities.
  *
- * Copyright (c) 2002-2011, PostgreSQL Global Development Group
+ * Copyright (c) 2002-2012, PostgreSQL Global Development Group
  *
  * IDENTIFICATION
  *		src/backend/utils/adt/lockfuncs.c
@@ -19,9 +19,10 @@
 #include "pgxc/pgxc.h"
 #include "pgxc/pgxcnode.h"
 #include "pgxc/nodemgr.h"
+#include "executor/spi.h"
+#include "tcop/utility.h"
 #endif
 #include "storage/predicate_internals.h"
-#include "storage/proc.h"
 #include "utils/builtins.h"
 
 
@@ -75,6 +76,9 @@ pgxc_advisory_lock(int64 key64, int32 key1, int32 key2, bool iskeybig,
 			LOCKMODE lockmode, LockLevel locklevel, TryType try);
 #endif
 
+/* Number of columns in pg_locks output */
+#define NUM_LOCK_STATUS_COLUMNS		15
+
 /*
  * VXIDGetDatum - Construct a text representation of a VXID
  *
@@ -121,7 +125,7 @@ pg_lock_status(PG_FUNCTION_ARGS)
 
 		/* build tupdesc for result tuples */
 		/* this had better match pg_locks view in system_views.sql */
-		tupdesc = CreateTemplateTupleDesc(14, false);
+		tupdesc = CreateTemplateTupleDesc(NUM_LOCK_STATUS_COLUMNS, false);
 		TupleDescInitEntry(tupdesc, (AttrNumber) 1, "locktype",
 						   TEXTOID, -1, 0);
 		TupleDescInitEntry(tupdesc, (AttrNumber) 2, "database",
@@ -150,6 +154,8 @@ pg_lock_status(PG_FUNCTION_ARGS)
 						   TEXTOID, -1, 0);
 		TupleDescInitEntry(tupdesc, (AttrNumber) 14, "granted",
 						   BOOLOID, -1, 0);
+		TupleDescInitEntry(tupdesc, (AttrNumber) 15, "fastpath",
+						   BOOLOID, -1, 0);
 
 		funcctx->tuple_desc = BlessTupleDesc(tupdesc);
 
@@ -174,21 +180,17 @@ pg_lock_status(PG_FUNCTION_ARGS)
 
 	while (mystatus->currIdx < lockData->nelements)
 	{
-		PROCLOCK   *proclock;
-		LOCK	   *lock;
-		PGPROC	   *proc;
 		bool		granted;
 		LOCKMODE	mode = 0;
 		const char *locktypename;
 		char		tnbuf[32];
-		Datum		values[14];
-		bool		nulls[14];
+		Datum		values[NUM_LOCK_STATUS_COLUMNS];
+		bool		nulls[NUM_LOCK_STATUS_COLUMNS];
 		HeapTuple	tuple;
 		Datum		result;
+		LockInstanceData *instance;
 
-		proclock = &(lockData->proclocks[mystatus->currIdx]);
-		lock = &(lockData->locks[mystatus->currIdx]);
-		proc = &(lockData->procs[mystatus->currIdx]);
+		instance = &(lockData->locks[mystatus->currIdx]);
 
 		/*
 		 * Look to see if there are any held lock modes in this PROCLOCK. If
@@ -196,14 +198,14 @@ pg_lock_status(PG_FUNCTION_ARGS)
 		 * again.
 		 */
 		granted = false;
-		if (proclock->holdMask)
+		if (instance->holdMask)
 		{
 			for (mode = 0; mode < MAX_LOCKMODES; mode++)
 			{
-				if (proclock->holdMask & LOCKBIT_ON(mode))
+				if (instance->holdMask & LOCKBIT_ON(mode))
 				{
 					granted = true;
-					proclock->holdMask &= LOCKBIT_OFF(mode);
+					instance->holdMask &= LOCKBIT_OFF(mode);
 					break;
 				}
 			}
@@ -215,10 +217,10 @@ pg_lock_status(PG_FUNCTION_ARGS)
 		 */
 		if (!granted)
 		{
-			if (proc->waitLock == proclock->tag.myLock)
+			if (instance->waitLockMode != NoLock)
 			{
 				/* Yes, so report it with proper mode */
-				mode = proc->waitLockMode;
+				mode = instance->waitLockMode;
 
 				/*
 				 * We are now done with this PROCLOCK, so advance pointer to
@@ -243,22 +245,22 @@ pg_lock_status(PG_FUNCTION_ARGS)
 		MemSet(values, 0, sizeof(values));
 		MemSet(nulls, false, sizeof(nulls));
 
-		if (lock->tag.locktag_type <= LOCKTAG_LAST_TYPE)
-			locktypename = LockTagTypeNames[lock->tag.locktag_type];
+		if (instance->locktag.locktag_type <= LOCKTAG_LAST_TYPE)
+			locktypename = LockTagTypeNames[instance->locktag.locktag_type];
 		else
 		{
 			snprintf(tnbuf, sizeof(tnbuf), "unknown %d",
-					 (int) lock->tag.locktag_type);
+					 (int) instance->locktag.locktag_type);
 			locktypename = tnbuf;
 		}
 		values[0] = CStringGetTextDatum(locktypename);
 
-		switch ((LockTagType) lock->tag.locktag_type)
+		switch ((LockTagType) instance->locktag.locktag_type)
 		{
 			case LOCKTAG_RELATION:
 			case LOCKTAG_RELATION_EXTEND:
-				values[1] = ObjectIdGetDatum(lock->tag.locktag_field1);
-				values[2] = ObjectIdGetDatum(lock->tag.locktag_field2);
+				values[1] = ObjectIdGetDatum(instance->locktag.locktag_field1);
+				values[2] = ObjectIdGetDatum(instance->locktag.locktag_field2);
 				nulls[3] = true;
 				nulls[4] = true;
 				nulls[5] = true;
@@ -268,9 +270,9 @@ pg_lock_status(PG_FUNCTION_ARGS)
 				nulls[9] = true;
 				break;
 			case LOCKTAG_PAGE:
-				values[1] = ObjectIdGetDatum(lock->tag.locktag_field1);
-				values[2] = ObjectIdGetDatum(lock->tag.locktag_field2);
-				values[3] = UInt32GetDatum(lock->tag.locktag_field3);
+				values[1] = ObjectIdGetDatum(instance->locktag.locktag_field1);
+				values[2] = ObjectIdGetDatum(instance->locktag.locktag_field2);
+				values[3] = UInt32GetDatum(instance->locktag.locktag_field3);
 				nulls[4] = true;
 				nulls[5] = true;
 				nulls[6] = true;
@@ -279,10 +281,10 @@ pg_lock_status(PG_FUNCTION_ARGS)
 				nulls[9] = true;
 				break;
 			case LOCKTAG_TUPLE:
-				values[1] = ObjectIdGetDatum(lock->tag.locktag_field1);
-				values[2] = ObjectIdGetDatum(lock->tag.locktag_field2);
-				values[3] = UInt32GetDatum(lock->tag.locktag_field3);
-				values[4] = UInt16GetDatum(lock->tag.locktag_field4);
+				values[1] = ObjectIdGetDatum(instance->locktag.locktag_field1);
+				values[2] = ObjectIdGetDatum(instance->locktag.locktag_field2);
+				values[3] = UInt32GetDatum(instance->locktag.locktag_field3);
+				values[4] = UInt16GetDatum(instance->locktag.locktag_field4);
 				nulls[5] = true;
 				nulls[6] = true;
 				nulls[7] = true;
@@ -290,7 +292,8 @@ pg_lock_status(PG_FUNCTION_ARGS)
 				nulls[9] = true;
 				break;
 			case LOCKTAG_TRANSACTION:
-				values[6] = TransactionIdGetDatum(lock->tag.locktag_field1);
+				values[6] =
+					TransactionIdGetDatum(instance->locktag.locktag_field1);
 				nulls[1] = true;
 				nulls[2] = true;
 				nulls[3] = true;
@@ -301,8 +304,8 @@ pg_lock_status(PG_FUNCTION_ARGS)
 				nulls[9] = true;
 				break;
 			case LOCKTAG_VIRTUALTRANSACTION:
-				values[5] = VXIDGetDatum(lock->tag.locktag_field1,
-										 lock->tag.locktag_field2);
+				values[5] = VXIDGetDatum(instance->locktag.locktag_field1,
+										 instance->locktag.locktag_field2);
 				nulls[1] = true;
 				nulls[2] = true;
 				nulls[3] = true;
@@ -316,10 +319,10 @@ pg_lock_status(PG_FUNCTION_ARGS)
 			case LOCKTAG_USERLOCK:
 			case LOCKTAG_ADVISORY:
 			default:			/* treat unknown locktags like OBJECT */
-				values[1] = ObjectIdGetDatum(lock->tag.locktag_field1);
-				values[7] = ObjectIdGetDatum(lock->tag.locktag_field2);
-				values[8] = ObjectIdGetDatum(lock->tag.locktag_field3);
-				values[9] = Int16GetDatum(lock->tag.locktag_field4);
+				values[1] = ObjectIdGetDatum(instance->locktag.locktag_field1);
+				values[7] = ObjectIdGetDatum(instance->locktag.locktag_field2);
+				values[8] = ObjectIdGetDatum(instance->locktag.locktag_field3);
+				values[9] = Int16GetDatum(instance->locktag.locktag_field4);
 				nulls[2] = true;
 				nulls[3] = true;
 				nulls[4] = true;
@@ -328,13 +331,14 @@ pg_lock_status(PG_FUNCTION_ARGS)
 				break;
 		}
 
-		values[10] = VXIDGetDatum(proc->backendId, proc->lxid);
-		if (proc->pid != 0)
-			values[11] = Int32GetDatum(proc->pid);
+		values[10] = VXIDGetDatum(instance->backend, instance->lxid);
+		if (instance->pid != 0)
+			values[11] = Int32GetDatum(instance->pid);
 		else
 			nulls[11] = true;
-		values[12] = CStringGetTextDatum(GetLockmodeName(LOCK_LOCKMETHOD(*lock), mode));
+		values[12] = CStringGetTextDatum(GetLockmodeName(instance->locktag.locktag_lockmethodid, mode));
 		values[13] = BoolGetDatum(granted);
+		values[14] = BoolGetDatum(instance->fastpath);
 
 		tuple = heap_form_tuple(funcctx->tuple_desc, values, nulls);
 		result = HeapTupleGetDatum(tuple);
@@ -352,8 +356,8 @@ pg_lock_status(PG_FUNCTION_ARGS)
 
 		PREDICATELOCKTARGETTAG *predTag = &(predLockData->locktags[mystatus->predLockIdx]);
 		SERIALIZABLEXACT *xact = &(predLockData->xacts[mystatus->predLockIdx]);
-		Datum		values[14];
-		bool		nulls[14];
+		Datum		values[NUM_LOCK_STATUS_COLUMNS];
+		bool		nulls[NUM_LOCK_STATUS_COLUMNS];
 		HeapTuple	tuple;
 		Datum		result;
 
@@ -400,10 +404,11 @@ pg_lock_status(PG_FUNCTION_ARGS)
 
 		/*
 		 * Lock mode. Currently all predicate locks are SIReadLocks, which are
-		 * always held (never waiting)
+		 * always held (never waiting) and have no fast path
 		 */
 		values[12] = CStringGetTextDatum("SIReadLock");
 		values[13] = BoolGetDatum(true);
+		values[14] = BoolGetDatum(false);
 
 		tuple = heap_form_tuple(funcctx->tuple_desc, values, nulls);
 		result = HeapTupleGetDatum(tuple);
@@ -479,8 +484,10 @@ pgxc_advisory_lock(int64 key64, int32 key1, int32 key2, bool iskeybig,
 	/* Skip everything XC specific if there's only one Coordinator running */
 	if (numcoords <= 1)
 	{
-		(void) LockAcquire(&locktag, lockmode, sessionLock, dontWait);
-		return true;
+		LockAcquireResult res;
+
+		res = LockAcquire(&locktag, lockmode, sessionLock, dontWait);
+		return (res == LOCKACQUIRE_OK || res == LOCKACQUIRE_ALREADY_HELD);
 	}
 
 	/*
@@ -1058,3 +1065,128 @@ pg_advisory_unlock_all(PG_FUNCTION_ARGS)
 
 	PG_RETURN_VOID();
 }
+
+#ifdef PGXC
+/*
+ * pgxc_lock_for_backup
+ *
+ * Lock the cluster for taking backup
+ * To lock the cluster, try to acquire a session level advisory lock exclusivly
+ * By lock we mean to disallow any statements that change
+ * the portions of the catalog which are backed up by pg_dump/pg_dumpall
+ * Returns true or fails with an error message.
+ */
+Datum
+pgxc_lock_for_backup(PG_FUNCTION_ARGS)
+{
+	bool lockAcquired = false;
+	int prepared_xact_count = 0;
+
+	if (!superuser())
+		ereport(ERROR,
+				(errcode(ERRCODE_INSUFFICIENT_PRIVILEGE),
+					errmsg("only superuser can lock the cluster for backup")));
+
+	/*
+	 * The system cannot be locked for backup if there is an uncommitted
+	 * prepared transaction, the reason is as follows:
+	 * Utility statements are divided into two groups, one is allowed group
+	 * and the other is disallowed group. A statement is put in allowed group
+	 * if it does not make changes to the catalog or makes such changes which
+	 * are not backed up by pg_dump or pg_dumpall, otherwise it is put in
+	 * disallowed group. Every time a disallowed statement is issued we try to
+	 * hold an advisory lock in shared mode and if the lock can be acquired
+	 * only then the statement is allowed.
+	 * In case of prepared transactions suppose the lock is not released at
+	 * prepare transaction 'txn_id'
+	 * Consider the following scenario:
+	 *
+	 *	begin;
+	 *	create table abc_def(a int, b int);
+	 *	insert into abc_def values(1,2),(3,4);
+	 *	prepare transaction 'abc';
+	 *
+	 * Now assume that the server is restarted for any reason.
+	 * When prepared transactions are saved on disk, session level locks are
+	 * ignored and hence when the prepared transactions are reterieved and all
+	 * the other locks are reclaimed, but session level advisory locks are
+	 * not reclaimed.
+	 * Hence we made the following decisions
+	 * a) Transaction level advisory locks should be used for DDLs which are
+	 *    automatically released at prepare transaction 'txn_id'
+	 * b) If there is any uncomitted prepared transaction, it is assumed
+	 *    that it must be issuing a statement that belongs to disallowed
+	 *    group and hence the request to hold the advisory lock exclusively
+	 *    is denied.
+	 */
+
+	/* Connect to SPI manager to check any prepared transactions */
+	if (SPI_connect() < 0)
+	{
+		ereport(ERROR,
+				(errcode(ERRCODE_CONNECTION_EXCEPTION),
+					errmsg("internal error while locking the cluster for backup")));
+	}
+
+	/* Are there any prepared transactions that have not yet been committed? */
+	SPI_execute("select gid from pg_catalog.pg_prepared_xacts limit 1", true, 0);
+	prepared_xact_count = SPI_processed;
+	SPI_finish();
+
+	if (prepared_xact_count > 0)
+	{
+		ereport(ERROR,
+				(errcode(ERRCODE_INTERNAL_ERROR),
+				errmsg("cannot lock cluster for backup in presence of %d uncommitted prepared transactions",
+						prepared_xact_count)));
+	}
+
+	/* try to acquire the advisory lock in exclusive mode */
+	lockAcquired = DatumGetBool(DirectFunctionCall2(
+										pg_try_advisory_lock_int4,
+										xc_lockForBackupKey1,
+										xc_lockForBackupKey2));
+
+	if (!lockAcquired)
+		ereport(ERROR,
+				(errcode(ERRCODE_LOCK_NOT_AVAILABLE),
+				errmsg("cannot lock cluster for backup, lock is already held")));
+
+	/*
+	 * sessin level advisory locks stay for only as long as the session
+	 * that issues them does
+	 */
+	elog(INFO, "please do not close this session until you are done adding the new node");
+
+	/* will be true always */
+	PG_RETURN_BOOL(lockAcquired);
+}
+
+/*
+ * pgxc_lock_for_backup
+ *
+ * Lock the cluster for taking backup
+ * To lock the cluster, try to acquire a session level advisory lock exclusivly
+ * By lock we mean to disallow any statements that change
+ * the portions of the catalog which are backed up by pg_dump/pg_dumpall
+ * Returns true or fails with an error message.
+ */
+bool
+pgxc_lock_for_utility_stmt(Node *parsetree)
+{
+	bool lockAcquired;
+
+	lockAcquired = DatumGetBool(DirectFunctionCall2(
+								pg_try_advisory_xact_lock_shared_int4,
+								xc_lockForBackupKey1,
+								xc_lockForBackupKey2));
+
+	if (!lockAcquired)
+		ereport(ERROR,
+				(errcode(ERRCODE_READ_ONLY_SQL_TRANSACTION),
+				errmsg("cannot execute %s in a locked cluster",
+						CreateCommandTag(parsetree))));
+
+	return lockAcquired;
+}
+#endif
