@@ -5,7 +5,7 @@
  *
  * Portions Copyright (c) 1996-2009, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
- * Portions Copyright (c) 2010-2013, Postgres-XC Development Group
+ * Portions Copyright (c) 2010-2012 Postgres-XC Development Group
  *
  *
  * IDENTIFICATION
@@ -570,7 +570,7 @@ gtmpqParseSuccess(GTM_Conn *conn, GTM_Result *result)
 			}
 
 			result->gr_resdata.grd_seq_list.seq =
-						(GTM_SeqInfo **)malloc(sizeof(GTM_SeqInfo *) *
+						(GTM_SeqInfo **)malloc(sizeof(GTM_SeqInfo) *
 											   result->gr_resdata.grd_seq_list.seq_count);
 
 			for (i = 0 ; i < result->gr_resdata.grd_seq_list.seq_count; i++)
@@ -593,7 +593,8 @@ gtmpqParseSuccess(GTM_Conn *conn, GTM_Result *result)
 					break;
 				}
 
-				result->gr_resdata.grd_seq_list.seq[i] = gtm_deserialize_sequence(buf, buflen);
+				gtm_deserialize_sequence(result->gr_resdata.grd_seq_list.seq+i,
+										 buf, buflen);
 
 				free(buf);
 			}
@@ -754,6 +755,8 @@ gtmpqParseSuccess(GTM_Conn *conn, GTM_Result *result)
 
 			break;
 		}
+		case BARRIER_RESULT:
+			break;
 		default:
 			printfGTMPQExpBuffer(&conn->errorMessage,
 							  "unexpected result type from server; result typr was \"%d\"\n",
@@ -826,6 +829,9 @@ gtmpqFreeResultData(GTM_Result *result, GTM_PGXCNodeType remote_type)
 		case TXN_GET_ALL_PREPARED_RESULT:
 			break;
 
+		case BARRIER_RESULT:
+			break;
+
 		case SNAPSHOT_GET_RESULT:
 		case SNAPSHOT_GXID_GET_RESULT:
 			/*
