@@ -5,7 +5,7 @@
  *
  * Portions Copyright (c) 1996-2009, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
- * Portions Copyright (c) 2010-2013, Postgres-XC Development Group
+ * Portions Copyright (c) 2010-2012 Postgres-XC Development Group
  *
  *
  * IDENTIFICATION
@@ -26,6 +26,7 @@
 #include "gtm/libpq.h"
 #include "gtm/libpq-int.h"
 #include "gtm/pqformat.h"
+#include "gtm/gtm_backup.h"
 
 extern bool Backup_synchronously;
 
@@ -35,9 +36,6 @@ typedef struct GTM_SeqInfoHashBucket
 	GTM_RWLock	shb_lock;
 } GTM_SeqInfoHashBucket;
 
-static int SeqStartMagic =	0xfafafafa;
-static int SeqEndMagic =	0xfefefefe;
-
 #define SEQ_HASH_TABLE_SIZE		1024
 static GTM_SeqInfoHashBucket GTMSequences[SEQ_HASH_TABLE_SIZE];
 
@@ -50,6 +48,8 @@ static int seq_add_seqinfo(GTM_SeqInfo *seqinfo);
 static int seq_remove_seqinfo(GTM_SeqInfo *seqinfo);
 static GTM_SequenceKey seq_copy_key(GTM_SequenceKey key);
 static int seq_drop_with_dbkey(GTM_SequenceKey nsp);
+static bool GTM_NeedSeqRestoreUpdateInternal(GTM_SeqInfo *seqinfo);
+
 
 /*
  * Get the hash value given the sequence key
@@ -337,12 +337,16 @@ GTM_SeqOpen(GTM_SequenceKey seqkey,
 	/* Set the last value in case of a future restart */
 	seqinfo->gs_last_value = seqinfo->gs_init_value;
 
+	seqinfo->gs_backedUpValue = seqinfo->gs_value;
+
 	if ((errcode = seq_add_seqinfo(seqinfo)))
 	{
 		GTM_RWLockDestroy(&seqinfo->gs_lock);
 		pfree(seqinfo->gs_key);
 		pfree(seqinfo);
 	}
+	GTM_SetNeedBackup();
+
 	return errcode;
 }
 
@@ -438,6 +442,7 @@ GTM_SeqRestore(GTM_SequenceKey seqkey,
 
 	seqinfo->gs_init_value = seqinfo->gs_last_value = startval;
 	seqinfo->gs_value = curval;
+	seqinfo->gs_backedUpValue = seqinfo->gs_value;
 
 	/*
 	 * Should we wrap around ?
@@ -625,6 +630,7 @@ GTM_SeqRename(GTM_SequenceKey seqkey, GTM_SequenceKey newseqkey)
 
 	newseqinfo->gs_init_value = seqinfo->gs_init_value;
 	newseqinfo->gs_value = seqinfo->gs_value;
+	newseqinfo->gs_backedUpValue = seqinfo->gs_backedUpValue;
 	newseqinfo->gs_cycle = seqinfo->gs_cycle;
 
 	newseqinfo->gs_state = seqinfo->gs_state;
@@ -684,6 +690,7 @@ GTM_SeqSetVal(GTM_SequenceKey seqkey, GTM_Sequence nextval, bool iscalled)
 
 	/* Remove the old key with the old name */
 	GTM_RWLockRelease(&seqinfo->gs_lock);
+	GTM_SetNeedBackup();
 	seq_release_seqinfo(seqinfo);
 
 	return 0;
@@ -769,6 +776,8 @@ GTM_SeqGetNext(GTM_SequenceKey seqkey)
 
 	}
 	GTM_RWLockRelease(&seqinfo->gs_lock);
+	if (GTM_NeedSeqRestoreUpdateInternal(seqinfo))
+		GTM_SetNeedBackup();
 	seq_release_seqinfo(seqinfo);
 	return value;
 }
@@ -790,10 +799,12 @@ GTM_SeqReset(GTM_SequenceKey seqkey)
 	}
 
 	GTM_RWLockAcquire(&seqinfo->gs_lock, GTM_LOCKMODE_WRITE);
-	seqinfo->gs_value = seqinfo->gs_last_value = seqinfo->gs_init_value;
+	seqinfo->gs_value = seqinfo->gs_last_value = seqinfo->gs_backedUpValue = seqinfo->gs_init_value;
 	GTM_RWLockRelease(&seqinfo->gs_lock);
 
+	GTM_SetNeedBackup();
 	seq_release_seqinfo(seqinfo);
+
 	return 0;
 }
 
@@ -872,7 +883,7 @@ ProcessSequenceInitCommand(Port *myport, StringInfo message, bool is_backup)
 			GTM_Conn *oldconn = GetMyThreadInfo->thr_conn->standby;
 			int count = 0;
 
-			elog(LOG, "calling open_sequence() for standby GTM %p.", GetMyThreadInfo->thr_conn->standby);
+			elog(DEBUG1, "calling open_sequence() for standby GTM %p.", GetMyThreadInfo->thr_conn->standby);
 
 		retry:
 			rc = bkup_open_sequence(GetMyThreadInfo->thr_conn->standby,
@@ -890,7 +901,7 @@ ProcessSequenceInitCommand(Port *myport, StringInfo message, bool is_backup)
 			if (Backup_synchronously && (myport->remote_type != GTM_NODE_GTM_PROXY))
 				gtm_sync_standby(GetMyThreadInfo->thr_conn->standby);
 
-			elog(LOG, "open_sequence() returns rc %d.", rc);
+			elog(DEBUG1, "open_sequence() returns rc %d.", rc);
 		}
 		/*
 		 * Send a SUCCESS message back to the client
@@ -984,7 +995,7 @@ ProcessSequenceAlterCommand(Port *myport, StringInfo message, bool is_backup)
 			GTM_Conn *oldconn = GetMyThreadInfo->thr_conn->standby;
 			int count = 0;
 
-			elog(LOG, "calling alter_sequence() for standby GTM %p.", GetMyThreadInfo->thr_conn->standby);
+			elog(DEBUG1, "calling alter_sequence() for standby GTM %p.", GetMyThreadInfo->thr_conn->standby);
 
 		retry:
 			rc = bkup_alter_sequence(GetMyThreadInfo->thr_conn->standby,
@@ -1004,7 +1015,7 @@ ProcessSequenceAlterCommand(Port *myport, StringInfo message, bool is_backup)
 			if (Backup_synchronously && (myport->remote_type != GTM_NODE_GTM_PROXY))
 				gtm_sync_standby(GetMyThreadInfo->thr_conn->standby);
 
-			elog(LOG, "alter_sequence() returns rc %d.", rc);
+			elog(DEBUG1, "alter_sequence() returns rc %d.", rc);
 		}
 		pq_beginmessage(&buf, 'S');
 		pq_sendint(&buf, SEQUENCE_ALTER_RESULT, 4);
@@ -1037,9 +1048,9 @@ void
 ProcessSequenceListCommand(Port *myport, StringInfo message)
 {
 	StringInfoData buf;
-	int seq_count = 0;
-	MemoryContext oldContext;
-	GTM_SeqInfo *seq_list[1024]; /* FIXME: make it expandable. */
+	int seq_count;
+	int seq_maxcount;
+	GTM_SeqInfo **seq_list;
 	int i;
 
 	if (Recovery_IsStandby())
@@ -1047,39 +1058,57 @@ ProcessSequenceListCommand(Port *myport, StringInfo message)
 			(EPERM,
 			 errmsg("Operation not permitted under the standby mode.")));
 
-	memset(seq_list, 0, sizeof(GTM_SeqInfo *) * 1024);
-
-	/*
-	 * We must use the TopMostMemoryContext because the sequence information is
-	 * not bound to a thread and can outlive any of the thread specific
-	 * contextes.
-	 */
-	oldContext = MemoryContextSwitchTo(TopMostMemoryContext);
+	seq_count = 0;
+	seq_maxcount = 1024;
+	seq_list = (GTM_SeqInfo **) palloc(seq_maxcount * sizeof(GTM_SeqInfo *));;
 
 	/*
 	 * Store pointers to all GTM_SeqInfo in the hash buckets into an array.
 	 */
+	for (i = 0 ; i < SEQ_HASH_TABLE_SIZE ; i++)
 	{
 		GTM_SeqInfoHashBucket *b;
 		gtm_ListCell *elem;
 
-		for (i = 0 ; i < SEQ_HASH_TABLE_SIZE ; i++)
+		b = &GTMSequences[i];
+
+		GTM_RWLockAcquire(&b->shb_lock, GTM_LOCKMODE_READ);
+
+		gtm_foreach(elem, b->shb_list)
 		{
-			b = &GTMSequences[i];
-
-			GTM_RWLockAcquire(&b->shb_lock, GTM_LOCKMODE_READ);
-
-			gtm_foreach(elem, b->shb_list)
+			/* Allocate larger array if required */
+			if (seq_count == seq_maxcount)
 			{
-				seq_list[seq_count] = (GTM_SeqInfo *) gtm_lfirst(elem);
-				seq_count++;
+				int 			newcount;
+				GTM_SeqInfo   **newlist;
+
+				newcount = 2 * seq_maxcount;
+				newlist = (GTM_SeqInfo **) repalloc(seq_list, newcount * sizeof(GTM_SeqInfo *));
+				/*
+				 * If failed try to get less. It is unlikely to happen, but
+				 * let's be safe.
+				 */
+				while (newlist == NULL)
+				{
+					newcount = seq_maxcount + (newcount - seq_maxcount) / 2 - 1;
+					if (newcount <= seq_maxcount)
+					{
+						/* give up */
+						ereport(ERROR,
+								(ERANGE,
+								 errmsg("Can not list all the sequences")));
+					}
+					newlist = (GTM_SeqInfo **) repalloc(seq_list, newcount * sizeof(GTM_SeqInfo *));
+				}
+				seq_maxcount = newcount;
+				seq_list = newlist;
 			}
-
-			GTM_RWLockRelease(&b->shb_lock);
+			seq_list[seq_count] = (GTM_SeqInfo *) gtm_lfirst(elem);
+			seq_count++;
 		}
-	}
 
-	MemoryContextSwitchTo(oldContext);
+		GTM_RWLockRelease(&b->shb_lock);
+	}
 
 	pq_getmsgend(message);
 
@@ -1096,27 +1125,38 @@ ProcessSequenceListCommand(Port *myport, StringInfo message)
 	/* Send a number of sequences */
 	pq_sendint(&buf, seq_count, 4);
 
-	for (i = 0 ; i < seq_count ; i++)
+	/*
+	 * Send sequences from the array
+	 */
 	{
-		char *seq_buf;
-		size_t seq_buflen;
+		/*
+		 * TODO set initial size big enough to fit any sequence, and avoid
+		 * reallocations.
+		 */
+		size_t seq_maxlen = 256;
+		char *seq_buf = (char *) palloc(seq_maxlen);
 
-		seq_buflen = gtm_get_sequence_size(seq_list[i]);
-		seq_buf = (char *)malloc(seq_buflen);
+		for (i = 0 ; i < seq_count ; i++)
+		{
+			size_t seq_buflen = gtm_get_sequence_size(seq_list[i]);
+			if (seq_buflen > seq_maxlen)
+			{
+				seq_maxlen = seq_buflen;
+				seq_buf = (char *)repalloc(seq_buf, seq_maxlen);
+			}
 
-		gtm_serialize_sequence(seq_list[i], seq_buf, seq_buflen);
+			gtm_serialize_sequence(seq_list[i], seq_buf, seq_buflen);
 
-		elog(LOG, "seq_buflen = %ld", seq_buflen);
+			elog(DEBUG1, "seq_buflen = %ld", seq_buflen);
 
-		pq_sendint(&buf, seq_buflen, 4);
-		pq_sendbytes(&buf, seq_buf, seq_buflen);
-
-		free(seq_buf);
+			pq_sendint(&buf, seq_buflen, 4);
+			pq_sendbytes(&buf, seq_buf, seq_buflen);
+		}
 	}
 
 	pq_endmessage(myport, &buf);
 
-	elog(LOG, "ProcessSequenceListCommand() done.");
+	elog(DEBUG1, "ProcessSequenceListCommand() done.");
 
 	if (myport->remote_type != GTM_NODE_GTM_PROXY)
 		/* Don't flush to the backup because this does not change the internal status */
@@ -1145,7 +1185,7 @@ ProcessSequenceGetNextCommand(Port *myport, StringInfo message, bool is_backup)
 				(ERANGE,
 				 errmsg("Can not get current value of the sequence")));
 
-	elog(LOG, "Getting next value %ld for sequence %s", seqval, seqkey.gsk_key);
+	elog(DEBUG1, "Getting next value %ld for sequence %s", seqval, seqkey.gsk_key);
 
 	if (!is_backup)
 	{
@@ -1156,10 +1196,10 @@ ProcessSequenceGetNextCommand(Port *myport, StringInfo message, bool is_backup)
 			GTM_Conn *oldconn = GetMyThreadInfo->thr_conn->standby;
 			int count = 0;
 
-			elog(LOG, "calling get_next() for standby GTM %p.", GetMyThreadInfo->thr_conn->standby);
+			elog(DEBUG1, "calling get_next() for standby GTM %p.", GetMyThreadInfo->thr_conn->standby);
 
 		retry:
-			loc_seq = bkup_get_next(GetMyThreadInfo->thr_conn->standby, &seqkey);
+			bkup_get_next(GetMyThreadInfo->thr_conn->standby, &seqkey, &loc_seq);
 
 			if (gtm_standby_check_communication_error(&count, oldconn))
 				goto retry;
@@ -1168,7 +1208,7 @@ ProcessSequenceGetNextCommand(Port *myport, StringInfo message, bool is_backup)
 			if (Backup_synchronously &&(myport->remote_type != GTM_NODE_GTM_PROXY))
 				gtm_sync_standby(GetMyThreadInfo->thr_conn->standby);
 
-			elog(LOG, "get_next() returns GTM_Sequence %ld.", loc_seq);
+			elog(DEBUG1, "get_next() returns GTM_Sequence %ld.", loc_seq);
 		}
 		/* Respond to the client */
 		pq_beginmessage(&buf, 'S');
@@ -1230,7 +1270,7 @@ ProcessSequenceSetValCommand(Port *myport, StringInfo message, bool is_backup)
 	 */
 	oldContext = MemoryContextSwitchTo(TopMostMemoryContext);
 
-	elog(LOG, "Setting new value %ld for sequence %s", nextval, seqkey.gsk_key);
+	elog(DEBUG1, "Setting new value %ld for sequence %s", nextval, seqkey.gsk_key);
 
 	if ((errcode = GTM_SeqSetVal(&seqkey, nextval, iscalled)))
 		ereport(ERROR,
@@ -1250,7 +1290,7 @@ ProcessSequenceSetValCommand(Port *myport, StringInfo message, bool is_backup)
 			GTM_Conn *oldconn = GetMyThreadInfo->thr_conn->standby;
 			int count = 0;
 
-			elog(LOG, "calling set_val() for standby GTM %p.", GetMyThreadInfo->thr_conn->standby);
+			elog(DEBUG1, "calling set_val() for standby GTM %p.", GetMyThreadInfo->thr_conn->standby);
 
 		retry:
 			rc = bkup_set_val(GetMyThreadInfo->thr_conn->standby,
@@ -1265,7 +1305,7 @@ ProcessSequenceSetValCommand(Port *myport, StringInfo message, bool is_backup)
 			if (Backup_synchronously && (myport->remote_type != GTM_NODE_GTM_PROXY))
 				gtm_sync_standby(GetMyThreadInfo->thr_conn->standby);
 
-			elog(LOG, "set_val() returns rc %d.", rc);
+			elog(DEBUG1, "set_val() returns rc %d.", rc);
 		}
 		/* Respond to the client */
 		pq_beginmessage(&buf, 'S');
@@ -1323,7 +1363,7 @@ ProcessSequenceResetCommand(Port *myport, StringInfo message, bool is_backup)
 			GTM_Conn *oldconn = GetMyThreadInfo->thr_conn->standby;
 			int count = 0;
 
-			elog(LOG, "calling reset_sequence() for standby GTM %p.", GetMyThreadInfo->thr_conn->standby);
+			elog(DEBUG1, "calling reset_sequence() for standby GTM %p.", GetMyThreadInfo->thr_conn->standby);
 
 		retry:
 			rc = bkup_reset_sequence(GetMyThreadInfo->thr_conn->standby, &seqkey);
@@ -1335,7 +1375,7 @@ ProcessSequenceResetCommand(Port *myport, StringInfo message, bool is_backup)
 			if (Backup_synchronously && (myport->remote_type != GTM_NODE_GTM_PROXY))
 				gtm_sync_standby(GetMyThreadInfo->thr_conn->standby);
 
-			elog(LOG, "reset_sequence() returns rc %d.", rc);
+			elog(DEBUG1, "reset_sequence() returns rc %d.", rc);
 		}
 		/* Respond to the client */
 		pq_beginmessage(&buf, 'S');
@@ -1379,7 +1419,7 @@ ProcessSequenceCloseCommand(Port *myport, StringInfo message, bool is_backup)
 	memcpy(&seqkey.gsk_type, pq_getmsgbytes(message, sizeof (GTM_SequenceKeyType)),
 		   sizeof (GTM_SequenceKeyType));
 
-	elog(LOG, "Closing sequence %s", seqkey.gsk_key);
+	elog(DEBUG1, "Closing sequence %s", seqkey.gsk_key);
 
 	if ((errcode = GTM_SeqClose(&seqkey)))
 		ereport(ERROR,
@@ -1395,7 +1435,7 @@ ProcessSequenceCloseCommand(Port *myport, StringInfo message, bool is_backup)
 			GTM_Conn *oldconn = GetMyThreadInfo->thr_conn->standby;
 			int count = 0;
 
-			elog(LOG, "calling close_sequence() for standby GTM %p.", GetMyThreadInfo->thr_conn->standby);
+			elog(DEBUG1, "calling close_sequence() for standby GTM %p.", GetMyThreadInfo->thr_conn->standby);
 
 		retry:
 			rc = bkup_close_sequence(GetMyThreadInfo->thr_conn->standby, &seqkey);
@@ -1407,7 +1447,7 @@ ProcessSequenceCloseCommand(Port *myport, StringInfo message, bool is_backup)
 			if (Backup_synchronously && (myport->remote_type != GTM_NODE_GTM_PROXY))
 				gtm_sync_standby(GetMyThreadInfo->thr_conn->standby);
 
-			elog(LOG, "close_sequence() returns rc %d.", rc);
+			elog(DEBUG1, "close_sequence() returns rc %d.", rc);
 		}
 		/* Respond to the client */
 		pq_beginmessage(&buf, 'S');
@@ -1482,7 +1522,7 @@ ProcessSequenceRenameCommand(Port *myport, StringInfo message, bool is_backup)
 			GTM_Conn *oldconn = GetMyThreadInfo->thr_conn->standby;
 			int count = 0;
 
-			elog(LOG, "calling rename_sequence() for standby GTM %p.", GetMyThreadInfo->thr_conn->standby);
+			elog(DEBUG1, "calling rename_sequence() for standby GTM %p.", GetMyThreadInfo->thr_conn->standby);
 
 		retry:
 			rc = bkup_rename_sequence(GetMyThreadInfo->thr_conn->standby, &seqkey, &newseqkey);
@@ -1494,7 +1534,7 @@ ProcessSequenceRenameCommand(Port *myport, StringInfo message, bool is_backup)
 			if (Backup_synchronously && (myport->remote_type != GTM_NODE_GTM_PROXY))
 				gtm_sync_standby(GetMyThreadInfo->thr_conn->standby);
 
-			elog(LOG, "rename_sequence() returns rc %d.", rc);
+			elog(DEBUG1, "rename_sequence() returns rc %d.", rc);
 		}
 		/* Send a SUCCESS message back to the client */
 		pq_beginmessage(&buf, 'S');
@@ -1521,8 +1561,252 @@ ProcessSequenceRenameCommand(Port *myport, StringInfo message, bool is_backup)
 	/* FIXME: need to check errors */
 }
 
-void
-GTM_SaveSeqInfo(int ctlfd)
+
+
+/*
+ * Escape whitespace and non-printable characters in the sequence name to
+ * store it to the control file.
+ */
+static void
+encode_seq_key(GTM_SequenceKey seqkey, char *buffer)
+{
+	int 	i;
+	char	c;
+	char   *out;
+
+	out = buffer;
+	for (i = 0; i < seqkey->gsk_keylen; i++)
+	{
+		c = seqkey->gsk_key[i];
+
+		if (c == '\\') /* double backslach */
+		{
+			*out++ = '\\';
+			*out++ = '\\';
+		}
+		else if (c > ' ') /* no need to escape */
+		{
+			*out++ = c;
+		}
+		else if (c == '\n') /* below some known non-printable chars */
+		{
+			*out++ = '\\';
+			*out++ = 'n';
+		}
+		else if (c == '\r')
+		{
+			*out++ = '\\';
+			*out++ = 'r';
+		}
+		else if (c == '\t')
+		{
+			*out++ = '\\';
+			*out++ = 't';
+		}
+		else /* other non-printable chars */
+		{
+			*out++ = '\\';
+			if ((int) c < 10)
+			{
+				*out++ = '0';
+				*out++ = (char) ((int) '0' + (int) c);
+			}
+			else
+			{
+				*out++ = (char) ((int) '0' + ((int) c) / 10);
+				*out++ = (char) ((int) '0' + ((int) c) % 10);
+			}
+		}
+	}
+	/* Add NULL terminator */
+	*out++ = '\0';
+}
+
+
+/*
+ * Decode the string encoded by the encode_seq_key function
+ */
+static void
+decode_seq_key(char* value, GTM_SequenceKey seqkey)
+{
+	char   *in;
+	char 	out[1024];
+	int		len = 0;
+
+	in = value;
+	while (*in != '\0')
+	{
+		if (*in == '\\') /* get escaped character */
+		{
+			in++; /* next value */
+			if (*in == '\\')
+				out[len++] = *in++;
+			else if (*in == 'n')
+			{
+				out[len++] = '\n';
+				in++;
+			}
+			else if (*in == 'r')
+			{
+				out[len++] = '\r';
+				in++;
+			}
+			else if (*in == 't')
+			{
+				out[len++] = '\t';
+				in++;
+			}
+			else /* \nn format */
+			{
+				int val;
+				val = ((int) *in++ - (int) '0');
+				val *= 10;
+				val += ((int) *in++ - (int) '0');
+				out[len++] = (char) val;
+			}
+		}
+		else /* get plain character */
+		{
+			out[len++] = *in++;
+		}
+	}
+	/* copy result to palloc'ed memory */
+	seqkey->gsk_keylen = len;
+	seqkey->gsk_key = (char *) palloc(len);
+	memcpy(seqkey->gsk_key, out, len);
+}
+
+static GTM_Sequence distanceToBackedUpSeqValue(GTM_SeqInfo *seqinfo)
+{
+	if (!SEQ_IS_CALLED(seqinfo))
+		return (GTM_Sequence)0;
+	if (SEQ_IS_ASCENDING(seqinfo))
+	{
+		if ((seqinfo->gs_backedUpValue - seqinfo->gs_value) >= 0)
+			return (seqinfo->gs_backedUpValue - seqinfo->gs_value);
+		if (SEQ_IS_CYCLE(seqinfo))
+			return((seqinfo->gs_max_value - seqinfo->gs_value) + (seqinfo->gs_backedUpValue - seqinfo->gs_min_value));
+		else
+			return(seqinfo->gs_backedUpValue - seqinfo->gs_value);
+	}
+	else
+	{
+		if ((seqinfo->gs_value - seqinfo->gs_backedUpValue) >= 0)
+			return(seqinfo->gs_backedUpValue - seqinfo->gs_value);
+		if (SEQ_IS_CYCLE(seqinfo))
+			return((seqinfo->gs_max_value - seqinfo->gs_backedUpValue) + (seqinfo->gs_value - seqinfo->gs_min_value));
+		else
+			return(seqinfo->gs_backedUpValue - seqinfo->gs_value);
+	}
+	return 0;
+}
+
+bool GTM_NeedSeqRestoreUpdate(GTM_SequenceKey seqkey)
+{
+	GTM_SeqInfo *seqinfo = seq_find_seqinfo(seqkey);
+	if (!seqinfo)
+		return FALSE;
+	return GTM_NeedSeqRestoreUpdateInternal(seqinfo);
+}
+
+static bool GTM_NeedSeqRestoreUpdateInternal(GTM_SeqInfo *seqinfo)
+{
+	GTM_Sequence distance;
+
+	if (!SEQ_IS_CALLED(seqinfo))
+		/* The first call.  Must backup */
+		return TRUE;
+	distance = distanceToBackedUpSeqValue(seqinfo);
+	if (SEQ_IS_ASCENDING(seqinfo))
+		return(distance >= seqinfo->gs_increment_by);
+	else
+		return(distance <= seqinfo->gs_increment_by);
+}
+
+
+
+static void
+GTM_SaveSeqInfo2(FILE *ctlf, bool isBackup)
+{
+	GTM_SeqInfoHashBucket *bucket;
+	gtm_ListCell *elem;
+	GTM_SeqInfo *seqinfo = NULL;
+	int hash;
+	char buffer[1024];
+
+	for (hash = 0; hash < SEQ_HASH_TABLE_SIZE; hash++)
+	{
+		bucket = &GTMSequences[hash];
+
+		GTM_RWLockAcquire(&bucket->shb_lock, GTM_LOCKMODE_READ);
+
+		gtm_foreach(elem, bucket->shb_list)
+		{
+			seqinfo = (GTM_SeqInfo *) gtm_lfirst(elem);
+			if (seqinfo == NULL)
+				break;
+
+			if (seqinfo->gs_state == SEQ_STATE_DELETED)
+				continue;
+
+			GTM_RWLockAcquire(&seqinfo->gs_lock, GTM_LOCKMODE_READ);
+
+			encode_seq_key(seqinfo->gs_key, buffer);
+			fprintf(ctlf, "%s\t%ld\t%ld\t%ld\t%ld\t%ld\t%c\t%c\t%x\n",
+					buffer, isBackup ? seqinfo->gs_backedUpValue : seqinfo->gs_value,
+					seqinfo->gs_init_value, seqinfo->gs_increment_by,
+					seqinfo->gs_min_value, seqinfo->gs_max_value,
+					(seqinfo->gs_cycle ? 't' : 'f'),
+					(seqinfo->gs_called ? 't' : 'f'),
+					seqinfo->gs_state);
+
+				GTM_RWLockRelease(&seqinfo->gs_lock);
+		}
+		GTM_RWLockRelease(&bucket->shb_lock);
+	}
+
+}
+
+void GTM_SaveSeqInfo(FILE *ctlf)
+{
+	GTM_SaveSeqInfo2(ctlf, FALSE);
+}
+
+static void advance_gs_value(GTM_SeqInfo *seqinfo)
+{
+	
+	GTM_Sequence distance;
+
+	distance = seqinfo->gs_increment_by * RestoreDuration;
+	if (SEQ_IS_ASCENDING(seqinfo))
+	{
+		if ((seqinfo->gs_max_value - seqinfo->gs_value) >= distance)
+			seqinfo->gs_backedUpValue = seqinfo->gs_value + distance;
+		else
+		{
+			if (SEQ_IS_CYCLE(seqinfo))
+				seqinfo->gs_backedUpValue = seqinfo->gs_min_value + (distance - (seqinfo->gs_max_value - seqinfo->gs_value));
+			else
+				seqinfo->gs_backedUpValue = seqinfo->gs_max_value;
+		}
+	}
+	else
+	{
+		if ((seqinfo->gs_min_value - seqinfo->gs_value) >= distance)
+			seqinfo->gs_backedUpValue = seqinfo->gs_value + distance;
+		else
+		{
+			if (SEQ_IS_CYCLE(seqinfo))
+				seqinfo->gs_backedUpValue = seqinfo->gs_max_value + (distance - (seqinfo->gs_min_value - seqinfo->gs_value));
+			else
+				seqinfo->gs_backedUpValue = seqinfo->gs_min_value;
+		}
+	}
+}
+
+
+static void
+GTM_UpdateRestorePointSeq(void)
 {
 	GTM_SeqInfoHashBucket *bucket;
 	gtm_ListCell *elem;
@@ -1545,20 +1829,7 @@ GTM_SaveSeqInfo(int ctlfd)
 				continue;
 
 			GTM_RWLockAcquire(&seqinfo->gs_lock, GTM_LOCKMODE_READ);
-
-			write(ctlfd, &SeqStartMagic, sizeof (SeqStartMagic));
-			write(ctlfd, &seqinfo->gs_key->gsk_keylen, sizeof (uint32));
-			write(ctlfd, seqinfo->gs_key->gsk_key, seqinfo->gs_key->gsk_keylen);
-			write(ctlfd, &seqinfo->gs_value, sizeof (GTM_Sequence));
-			write(ctlfd, &seqinfo->gs_init_value, sizeof (GTM_Sequence));
-			write(ctlfd, &seqinfo->gs_increment_by, sizeof (GTM_Sequence));
-			write(ctlfd, &seqinfo->gs_min_value, sizeof (GTM_Sequence));
-			write(ctlfd, &seqinfo->gs_max_value, sizeof (GTM_Sequence));
-			write(ctlfd, &seqinfo->gs_cycle, sizeof (bool));
-			write(ctlfd, &seqinfo->gs_called, sizeof (bool));
-			write(ctlfd, &seqinfo->gs_state, sizeof (int32));
-			write(ctlfd, &SeqEndMagic, sizeof(SeqEndMagic));
-
+			advance_gs_value(seqinfo);
 			GTM_RWLockRelease(&seqinfo->gs_lock);
 		}
 
@@ -1567,15 +1838,22 @@ GTM_SaveSeqInfo(int ctlfd)
 
 }
 
-void
-GTM_RestoreSeqInfo(int ctlfd)
-{
-	int magic;
 
-	if (ctlfd == -1)
+void GTM_WriteRestorePointSeq(FILE *ctlf)
+{
+	GTM_UpdateRestorePointSeq();
+	GTM_SaveSeqInfo2(ctlf, TRUE);
+}
+
+void
+GTM_RestoreSeqInfo(FILE *ctlf)
+{
+	char seqname[1024];
+
+	if (ctlf == NULL)
 		return;
 
-	while (read(ctlfd, &magic, sizeof (SeqStartMagic)) == sizeof (SeqStartMagic))
+	while (fscanf(ctlf, "%s", seqname) == 1)
 	{
 		GTM_SequenceKeyData seqkey;
 		GTM_Sequence increment_by;
@@ -1586,38 +1864,58 @@ GTM_RestoreSeqInfo(int ctlfd)
 		int32 state;
 		bool cycle;
 		bool called;
+		char boolval[16];
 
-		if (magic != SeqStartMagic)
-		{
-			elog(LOG, "Start magic mismatch %x - %x", magic, SeqStartMagic);
-			break;
-		}
+		decode_seq_key(seqname, &seqkey);
 
-		if (read(ctlfd, &seqkey.gsk_keylen, sizeof (uint32)) != sizeof (uint32))
-		{
-			elog(LOG, "Failed to read keylen");
-			break;
-		}
-
-		seqkey.gsk_key = palloc(seqkey.gsk_keylen);
-		read(ctlfd, seqkey.gsk_key, seqkey.gsk_keylen);
-
-		read(ctlfd, &curval, sizeof (GTM_Sequence));
-		read(ctlfd, &startval, sizeof (GTM_Sequence));
-		read(ctlfd, &increment_by, sizeof (GTM_Sequence));
-		read(ctlfd, &minval, sizeof (GTM_Sequence));
-		read(ctlfd, &maxval, sizeof (GTM_Sequence));
-		read(ctlfd, &cycle, sizeof (bool));
-		read(ctlfd, &called, sizeof (bool));
-		read(ctlfd, &state, sizeof (int32));
-		read(ctlfd, &magic, sizeof(SeqEndMagic));
-
-		if (magic != SeqEndMagic)
+		if (fscanf(ctlf, "%ld", &curval) != 1)
+		{
+			elog(WARNING, "Corrupted control file");
+			return;
+		}
+		if (fscanf(ctlf, "%ld", &startval) != 1)
+		{
+			elog(WARNING, "Corrupted control file");
+			return;
+		}
+		if (fscanf(ctlf, "%ld", &increment_by) != 1)
+		{
+			elog(WARNING, "Corrupted control file");
+			return;
+		}
+		if (fscanf(ctlf, "%ld", &minval) != 1)
+		{
+			elog(WARNING, "Corrupted control file");
+			return;
+		}
+		if (fscanf(ctlf, "%ld", &maxval) != 1)
+		{
+			elog(WARNING, "Corrupted control file");
+			return;
+		}
+		if (fscanf(ctlf, "%s", boolval) == 1)
+		{
+			cycle = (*boolval == 't');
+		}
+		else
+		{
+			elog(WARNING, "Corrupted control file");
+			return;
+		}
+		if (fscanf(ctlf, "%s", boolval) == 1)
+		{
+			called = (*boolval == 't');
+		}
+		else
+		{
+			elog(WARNING, "Corrupted control file");
+			return;
+		}
+		if (fscanf(ctlf, "%x", &state) != 1)
 		{
 			elog(WARNING, "Corrupted control file");
 			return;
 		}
-
 		GTM_SeqRestore(&seqkey, increment_by, minval, maxval, startval, curval,
 					   state, cycle, called);
 	}
