@@ -5,7 +5,7 @@
  *
  * Portions Copyright (c) 1996-2009, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
- * Portions Copyright (c) 2010-2013, Postgres-XC Development Group
+ * Portions Copyright (c) 2010-2012 Postgres-XC Development Group
  *
  *
  * IDENTIFICATION
@@ -27,6 +27,7 @@
 #include "gtm/libpq.h"
 #include "gtm/libpq-int.h"
 #include "gtm/pqformat.h"
+#include "gtm/gtm_backup.h"
 
 extern bool Backup_synchronously;
 
@@ -39,6 +40,7 @@ static void init_GTM_TransactionInfo(GTM_TransactionInfo *gtm_txninfo,
 									 GTM_IsolationLevel isolevel,
 									 GTMProxy_ConnID connid,
 									 bool readonly);
+static void clean_GTM_TransactionInfo(GTM_TransactionInfo *gtm_txninfo);
 GTM_Transactions GTMTransactions;
 
 void
@@ -155,7 +157,12 @@ GTM_GXIDToHandle(GlobalTransactionId gxid)
 	if (gtm_txninfo != NULL)
 		return gtm_txninfo->gti_handle;
 	else
+	{
+		ereport(WARNING,
+				(ERANGE, errmsg("No transaction handle for gxid: %d",
+								gxid)));
 		return InvalidTransactionHandle;
+	}
 }
 
 /*
@@ -259,21 +266,7 @@ GTM_RemoveTransInfoMulti(GTM_TransactionInfo *gtm_txninfo[], int txn_count)
 		/*
 		 * Now mark the transaction as aborted and mark the structure as not-in-use
 		 */
-		gtm_txninfo[ii]->gti_state = GTM_TXN_ABORTED;
-		gtm_txninfo[ii]->gti_in_use = false;
-		gtm_txninfo[ii]->gti_snapshot_set = false;
-
-		/* Clean-up also structures that were used for prepared transactions */
-		if (gtm_txninfo[ii]->gti_gid)
-		{
-			pfree(gtm_txninfo[ii]->gti_gid);
-			gtm_txninfo[ii]->gti_gid = NULL;
-		}
-		if (gtm_txninfo[ii]->nodestring)
-		{
-			pfree(gtm_txninfo[ii]->nodestring);
-			gtm_txninfo[ii]->nodestring = NULL;
-		}
+		clean_GTM_TransactionInfo(gtm_txninfo[ii]);
 	}
 
 	GTM_RWLockRelease(&GTMTransactions.gt_TransArrayLock);
@@ -328,20 +321,7 @@ GTM_RemoveAllTransInfos(int backend_id)
 			/*
 			 * Now mark the transaction as aborted and mark the structure as not-in-use
 			 */
-			gtm_txninfo->gti_state = GTM_TXN_ABORTED;
-			gtm_txninfo->gti_in_use = false;
-			gtm_txninfo->gti_snapshot_set = false;
-
-			if (gtm_txninfo->gti_gid)
-			{
-				pfree(gtm_txninfo->gti_gid);
-				gtm_txninfo->gti_gid = NULL;
-			}
-			if (gtm_txninfo->nodestring)
-			{
-				pfree(gtm_txninfo->nodestring);
-				gtm_txninfo->nodestring = NULL;
-			}
+			clean_GTM_TransactionInfo(gtm_txninfo);
 
 			/* move to next cell in the list */
 			if (prev)
@@ -578,10 +558,12 @@ GTM_GetGlobalTransactionIdMulti(GTM_TransactionHandle handle[], int txn_count)
 		gtm_txninfo = GTM_HandleToTransactionInfo(handle[ii]);
 		Assert(gtm_txninfo);
 
-		elog(LOG, "Assigning new transaction ID = %d", xid);
+		elog(DEBUG1, "Assigning new transaction ID = %d", xid);
 		gtm_txninfo->gti_gxid = xid;
 	}
 
+	if (GTM_NeedXidRestoreUpdate())
+		GTM_SetNeedBackup();
 	GTM_RWLockRelease(&GTMTransactions.gt_XidGenLock);
 
 	return start_xid;
@@ -621,7 +603,11 @@ ReadNewGlobalTransactionId(void)
  * started. When the GTM is finally shutdown, the next to-be-assigned GXID is
  * stroed in the control file.
  *
- * XXX We don't yet handle any crash recovery. So if the GTM is shutdown
+ * XXX We don't yet handle any crash recovery. So if the GTM is no shutdown normally...
+ *
+ * This is handled by gtm_backup.c.  Anyway, because this function is to be called by
+ * GTM_RestoreTransactionId() and the backup will be performed afterwords,
+ * we don't care the new value of GTMTransactions.gt_nextXid here.
  */
 void
 SetNextGlobalTransactionId(GlobalTransactionId gxid)
@@ -659,6 +645,8 @@ GTM_BeginTransactionMulti(char *coord_name,
 	 */
 	oldContext = MemoryContextSwitchTo(TopMostMemoryContext);
 
+	GTM_RWLockAcquire(&GTMTransactions.gt_TransArrayLock, GTM_LOCKMODE_WRITE);
+
 	for (kk = 0; kk < txn_count; kk++)
 	{
 		int ii, jj, startslot;
@@ -667,8 +655,6 @@ GTM_BeginTransactionMulti(char *coord_name,
 		 * We had no cached slots. Now find a free slot in the transation array
 		 * and store the transaction info structure there
 		 */
-		GTM_RWLockAcquire(&GTMTransactions.gt_TransArrayLock, GTM_LOCKMODE_WRITE);
-
 		startslot = GTMTransactions.gt_lastslot + 1;
 		if (startslot >= GTM_MAX_GLOBAL_TRANSACTIONS)
 			startslot = 0;
@@ -737,7 +723,7 @@ init_GTM_TransactionInfo(GTM_TransactionInfo *gtm_txninfo,
 	gtm_txninfo->gti_gxid = InvalidGlobalTransactionId;
 	gtm_txninfo->gti_xmin = InvalidGlobalTransactionId;
 	gtm_txninfo->gti_state = GTM_TXN_STARTING;
-	gtm_txninfo->gti_coordname = pstrdup(coord_name);
+	gtm_txninfo->gti_coordname = (coord_name ? pstrdup(coord_name) : NULL);
 
 	gtm_txninfo->gti_isolevel = isolevel;
 	gtm_txninfo->gti_readonly = readonly;
@@ -753,6 +739,35 @@ init_GTM_TransactionInfo(GTM_TransactionInfo *gtm_txninfo,
 }
 
 
+/*
+ * Clean up the TransactionInfo slot and pfree all the palloc'ed memory,
+ * except txid array of the snapshot, which is reused.
+ */
+static void
+clean_GTM_TransactionInfo(GTM_TransactionInfo *gtm_txninfo)
+{
+	gtm_txninfo->gti_state = GTM_TXN_ABORTED;
+	gtm_txninfo->gti_in_use = false;
+	gtm_txninfo->gti_snapshot_set = false;
+
+	if (gtm_txninfo->gti_coordname)
+	{
+		pfree(gtm_txninfo->gti_coordname);
+		gtm_txninfo->gti_coordname = NULL;
+	}
+	if (gtm_txninfo->gti_gid)
+	{
+		pfree(gtm_txninfo->gti_gid);
+		gtm_txninfo->gti_gid = NULL;
+	}
+	if (gtm_txninfo->nodestring)
+	{
+		pfree(gtm_txninfo->nodestring);
+		gtm_txninfo->nodestring = NULL;
+	}
+}
+
+
 void
 GTM_BkupBeginTransactionMulti(char *coord_name,
 							  GTM_TransactionHandle *txn,
@@ -1169,7 +1184,7 @@ ProcessBeginTransactionGetGXIDCommand(Port *myport, StringInfo message)
 
 	MemoryContextSwitchTo(oldContext);
 
-	elog(LOG, "Sending transaction id %u", gxid);
+	elog(DEBUG1, "Sending transaction id %u", gxid);
 
 	/* Backup first */
 	if (GetMyThreadInfo->thr_conn->standby)
@@ -1177,7 +1192,7 @@ ProcessBeginTransactionGetGXIDCommand(Port *myport, StringInfo message)
 		GTM_Conn *oldconn = GetMyThreadInfo->thr_conn->standby;
 		int count = 0;
 
-		elog(LOG, "calling begin_transaction() for standby GTM %p.", GetMyThreadInfo->thr_conn->standby);
+		elog(DEBUG1, "calling begin_transaction() for standby GTM %p.", GetMyThreadInfo->thr_conn->standby);
 
 retry:
 		bkup_begin_transaction_gxid(GetMyThreadInfo->thr_conn->standby,
@@ -1247,7 +1262,8 @@ GTM_BkupBeginTransactionGetGXIDMulti(char *coord_name,
 		GTMTransactions.gt_lastslot = txn[ii];
 		gtm_txninfo->gti_gxid = gxid[ii];
 		/*
-		 * Advance next gxid
+		 * Advance next gxid -- because this is called at slave only, we don't care the restoration point
+		 * here.  Restoration point will be created at promotion.
 		 */
 		if (GlobalTransactionIdPrecedes(GTMTransactions.gt_nextXid, gxid[ii]))
 			GTMTransactions.gt_nextXid = gxid[ii] + 1;
@@ -1327,7 +1343,7 @@ ProcessBeginTransactionGetGXIDAutovacuumCommand(Port *myport, StringInfo message
 	GlobalTransactionId gxid;
 	MemoryContext oldContext;
 
-	elog(LOG, "Inside ProcessBeginTransactionGetGXIDAutovacuumCommand");
+	elog(DEBUG1, "Inside ProcessBeginTransactionGetGXIDAutovacuumCommand");
 
 	txn_isolation_level = pq_getmsgint(message, sizeof (GTM_IsolationLevel));
 	txn_read_only = pq_getmsgbyte(message);
@@ -1356,7 +1372,7 @@ ProcessBeginTransactionGetGXIDAutovacuumCommand(Port *myport, StringInfo message
 
 	MemoryContextSwitchTo(oldContext);
 
-	elog(LOG, "Sending transaction id %d", gxid);
+	elog(DEBUG1, "Sending transaction id %d", gxid);
 
 	/* Backup first */
 	if (GetMyThreadInfo->thr_conn->standby)
@@ -1365,7 +1381,7 @@ ProcessBeginTransactionGetGXIDAutovacuumCommand(Port *myport, StringInfo message
 		GTM_Conn *oldconn = GetMyThreadInfo->thr_conn->standby;
 		int count = 0;
 
-		elog(LOG, "calling begin_transaction_autovacuum() for standby GTM %p.",
+		elog(DEBUG1, "calling begin_transaction_autovacuum() for standby GTM %p.",
 			 GetMyThreadInfo->thr_conn->standby);
 
 	retry:
@@ -1379,7 +1395,7 @@ ProcessBeginTransactionGetGXIDAutovacuumCommand(Port *myport, StringInfo message
 		if (Backup_synchronously && (myport->remote_type != GTM_NODE_GTM_PROXY))
 			gtm_sync_standby(GetMyThreadInfo->thr_conn->standby);
 
-		elog(LOG, "begin_transaction_autovacuum() GXID=%d done.", _gxid);
+		elog(DEBUG1, "begin_transaction_autovacuum() GXID=%d done.", _gxid);
 	}
 	/* Respond to the client */
 	pq_beginmessage(&buf, 'S');
@@ -1439,9 +1455,9 @@ ProcessBeginTransactionGetGXIDCommandMulti(Port *myport, StringInfo message)
 	/*
 	 * Start a new transaction
 	 *
-	 * XXX Port should contain Coordinator name - replace "" with that
+	 * XXX Port should contain Coordinator name - replace NULL with that
 	 */
-	count = GTM_BeginTransactionMulti("", txn_isolation_level, txn_read_only, txn_connid,
+	count = GTM_BeginTransactionMulti(NULL, txn_isolation_level, txn_read_only, txn_connid,
 									  txn_count, txn);
 	if (count != txn_count)
 		ereport(ERROR,
@@ -1459,11 +1475,11 @@ ProcessBeginTransactionGetGXIDCommandMulti(Port *myport, StringInfo message)
 	/* GXID has been received, now it's time to get a GTM timestamp */
 	timestamp = GTM_TimestampGetCurrent();
 
-	end_gxid = start_gxid + txn_count;
+	end_gxid = start_gxid + (txn_count - 1);
 	if (end_gxid < start_gxid)
 		end_gxid += FirstNormalGlobalTransactionId;
 
-	elog(LOG, "Sending transaction ids from %u to %u", start_gxid, end_gxid);
+	elog(DEBUG1, "Sending transaction ids from %u to %u", start_gxid, end_gxid);
 
 	/* Backup first */
 	if (GetMyThreadInfo->thr_conn->standby)
@@ -1472,7 +1488,7 @@ ProcessBeginTransactionGetGXIDCommandMulti(Port *myport, StringInfo message)
 		GTM_Conn *oldconn = GetMyThreadInfo->thr_conn->standby;
 		int count = 0;
 
-		elog(LOG, "calling begin_transaction_multi() for standby GTM %p.",
+		elog(DEBUG1, "calling begin_transaction_multi() for standby GTM %p.",
 		     GetMyThreadInfo->thr_conn->standby);
 
 retry:
@@ -1491,7 +1507,7 @@ retry:
 		if (Backup_synchronously && (myport->remote_type != GTM_NODE_GTM_PROXY))
 			gtm_sync_standby(GetMyThreadInfo->thr_conn->standby);
 
-		elog(LOG, "begin_transaction_multi() rc=%d done.", _rc);
+		elog(DEBUG1, "begin_transaction_multi() rc=%d done.", _rc);
 	}
 	/* Respond to the client */
 	pq_beginmessage(&buf, 'S');
@@ -1589,7 +1605,7 @@ ProcessCommitTransactionCommand(Port *myport, StringInfo message, bool is_backup
 
 	oldContext = MemoryContextSwitchTo(TopMemoryContext);
 
-	elog(LOG, "Committing transaction id %u", gxid);
+	elog(DEBUG1, "Committing transaction id %u", gxid);
 
 	/*
 	 * Commit the transaction
@@ -1609,7 +1625,7 @@ ProcessCommitTransactionCommand(Port *myport, StringInfo message, bool is_backup
 			GTM_Conn *oldconn = GetMyThreadInfo->thr_conn->standby;
 			int count = 0;
 
-			elog(LOG, "calling commit_transaction() for standby GTM %p.", GetMyThreadInfo->thr_conn->standby);
+			elog(DEBUG1, "calling commit_transaction() for standby GTM %p.", GetMyThreadInfo->thr_conn->standby);
 
 		retry:
 			_rc = bkup_commit_transaction(GetMyThreadInfo->thr_conn->standby, gxid);
@@ -1621,7 +1637,7 @@ ProcessCommitTransactionCommand(Port *myport, StringInfo message, bool is_backup
 			if (Backup_synchronously && (myport->remote_type != GTM_NODE_GTM_PROXY))
 				gtm_sync_standby(GetMyThreadInfo->thr_conn->standby);
 
-			elog(LOG, "commit_transaction() rc=%d done.", _rc);
+			elog(DEBUG1, "commit_transaction() rc=%d done.", _rc);
 		}
 
 		pq_beginmessage(&buf, 'S');
@@ -1697,7 +1713,7 @@ ProcessCommitPreparedTransactionCommand(Port *myport, StringInfo message, bool i
 
 	oldContext = MemoryContextSwitchTo(TopMemoryContext);
 
-	elog(LOG, "Committing: prepared id %u and commit prepared id %u ", gxid[0], gxid[1]);
+	elog(DEBUG1, "Committing: prepared id %u and commit prepared id %u ", gxid[0], gxid[1]);
 
 	/*
 	 * Commit the prepared transaction.
@@ -1715,7 +1731,7 @@ ProcessCommitPreparedTransactionCommand(Port *myport, StringInfo message, bool i
 			GTM_Conn *oldconn = GetMyThreadInfo->thr_conn->standby;
 			int count = 0;
 
-			elog(LOG, "calling commit_prepared_transaction() for standby GTM %p.",
+			elog(DEBUG1, "calling commit_prepared_transaction() for standby GTM %p.",
 				 GetMyThreadInfo->thr_conn->standby);
 
 		retry:
@@ -1729,7 +1745,7 @@ ProcessCommitPreparedTransactionCommand(Port *myport, StringInfo message, bool i
 			if (Backup_synchronously && (myport->remote_type != GTM_NODE_GTM_PROXY))
 				gtm_sync_standby(GetMyThreadInfo->thr_conn->standby);
 
-			elog(LOG, "commit_prepared_transaction() rc=%d done.", _rc);
+			elog(DEBUG1, "commit_prepared_transaction() rc=%d done.", _rc);
 		}
 		/* Respond to the client */
 		pq_beginmessage(&buf, 'S');
@@ -1858,7 +1874,7 @@ ProcessGetGIDDataTransactionCommand(Port *myport, StringInfo message)
 		GTM_Conn *oldconn = GetMyThreadInfo->thr_conn->standby;
 		int count = 0;
 
-		elog(LOG, "calling get_gid_data() for standby GTM %p.",
+		elog(DEBUG1, "calling get_gid_data() for standby GTM %p.",
 			GetMyThreadInfo->thr_conn->standby);
 
 retry:
@@ -1872,7 +1888,7 @@ retry:
 		if (gtm_standby_check_communication_error(&count, oldconn))
 			goto retry;
 
-		elog(LOG, "get_gid_data() rc=%d done.", _rc);
+		elog(DEBUG1, "get_gid_data() rc=%d done.", _rc);
 	}
 #endif
 
@@ -1909,7 +1925,7 @@ ProcessGXIDListCommand(Port *myport, StringInfo message)
 
 	actlen = gtm_serialize_transactions(&GTMTransactions, data, estlen);
 
-	elog(LOG, "gtm_serialize_transactions: estlen=%ld, actlen=%ld", estlen, actlen);
+	elog(DEBUG1, "gtm_serialize_transactions: estlen=%ld, actlen=%ld", estlen, actlen);
 
 	GTM_RWLockRelease(&GTMTransactions.gt_XidGenLock);
 
@@ -1935,10 +1951,10 @@ ProcessGXIDListCommand(Port *myport, StringInfo message)
 	if (myport->remote_type != GTM_NODE_GTM_PROXY)
 	{
 		pq_flush(myport);
-		elog(LOG, "pq_flush()");
+		elog(DEBUG1, "pq_flush()");
 	}
 
-	elog(LOG, "ProcessGXIDListCommand() ok. %ld bytes sent. len=%d", actlen, buf.len);
+	elog(DEBUG1, "ProcessGXIDListCommand() ok. %ld bytes sent. len=%d", actlen, buf.len);
 	free(data);
 
 	return;
@@ -1986,7 +2002,7 @@ ProcessRollbackTransactionCommand(Port *myport, StringInfo message, bool is_back
 
 	oldContext = MemoryContextSwitchTo(TopMemoryContext);
 
-	elog(LOG, "Cancelling transaction id %u", gxid);
+	elog(DEBUG1, "Cancelling transaction id %u", gxid);
 
 	/*
 	 * Commit the transaction
@@ -2003,7 +2019,7 @@ ProcessRollbackTransactionCommand(Port *myport, StringInfo message, bool is_back
 			GTM_Conn *oldconn = GetMyThreadInfo->thr_conn->standby;
 			int count = 0;
 
-			elog(LOG, "calling abort_transaction() for standby GTM %p.", GetMyThreadInfo->thr_conn->standby);
+			elog(DEBUG1, "calling abort_transaction() for standby GTM %p.", GetMyThreadInfo->thr_conn->standby);
 
 		retry:
 			bkup_abort_transaction(GetMyThreadInfo->thr_conn->standby, gxid);
@@ -2015,7 +2031,7 @@ ProcessRollbackTransactionCommand(Port *myport, StringInfo message, bool is_back
 			if (Backup_synchronously && (myport->remote_type != GTM_NODE_GTM_PROXY))
 				gtm_sync_standby(GetMyThreadInfo->thr_conn->standby);
 
-			elog(LOG, "abort_transaction() GXID=%d done.", gxid);
+			elog(DEBUG1, "abort_transaction() GXID=%d done.", gxid);
 		}
 		/* Respond to the client */
 		pq_beginmessage(&buf, 'S');
@@ -2108,7 +2124,7 @@ ProcessCommitTransactionCommandMulti(Port *myport, StringInfo message, bool is_b
 			GTM_Conn *oldconn = GetMyThreadInfo->thr_conn->standby;
 			int count = 0;
 
-			elog(LOG, "calling commit_transaction_multi() for standby GTM %p.",
+			elog(DEBUG1, "calling commit_transaction_multi() for standby GTM %p.",
 				 GetMyThreadInfo->thr_conn->standby);
 
 		retry:
@@ -2120,7 +2136,7 @@ ProcessCommitTransactionCommandMulti(Port *myport, StringInfo message, bool is_b
 			if (Backup_synchronously && (myport->remote_type != GTM_NODE_GTM_PROXY))
 				gtm_sync_standby(GetMyThreadInfo->thr_conn->standby);
 
-			elog(LOG, "commit_transaction_multi() rc=%d done.", _rc);
+			elog(DEBUG1, "commit_transaction_multi() rc=%d done.", _rc);
 		}
 		/* Respond to the client */
 		pq_beginmessage(&buf, 'S');
@@ -2211,7 +2227,7 @@ ProcessRollbackTransactionCommandMulti(Port *myport, StringInfo message, bool is
 			GTM_Conn *oldconn = GetMyThreadInfo->thr_conn->standby;
 			int count = 0;
 
-			elog(LOG, "calling abort_transaction_multi() for standby GTM %p.",
+			elog(DEBUG1, "calling abort_transaction_multi() for standby GTM %p.",
 				 GetMyThreadInfo->thr_conn->standby);
 
 		retry:
@@ -2224,7 +2240,7 @@ ProcessRollbackTransactionCommandMulti(Port *myport, StringInfo message, bool is
 			if (Backup_synchronously &&(myport->remote_type != GTM_NODE_GTM_PROXY))
 				gtm_sync_standby(GetMyThreadInfo->thr_conn->standby);
 
-			elog(LOG, "abort_transaction_multi() rc=%d done.", _rc);
+			elog(DEBUG1, "abort_transaction_multi() rc=%d done.", _rc);
 		}
 		/* Respond to the client */
 		pq_beginmessage(&buf, 'S');
@@ -2325,7 +2341,7 @@ ProcessStartPreparedTransactionCommand(Port *myport, StringInfo message, bool is
 			GTM_Conn *oldconn = GetMyThreadInfo->thr_conn->standby;
 			int count = 0;
 
-			elog(LOG, "calling start_prepared_transaction() for standby GTM %p.",
+			elog(DEBUG1, "calling start_prepared_transaction() for standby GTM %p.",
 				 GetMyThreadInfo->thr_conn->standby);
 
 		retry:
@@ -2340,7 +2356,7 @@ ProcessStartPreparedTransactionCommand(Port *myport, StringInfo message, bool is
 			if (Backup_synchronously && (myport->remote_type != GTM_NODE_GTM_PROXY))
 				gtm_sync_standby(GetMyThreadInfo->thr_conn->standby);
 
-			elog(LOG, "start_prepared_transaction() rc=%d done.", _rc);
+			elog(DEBUG1, "start_prepared_transaction() rc=%d done.", _rc);
 		}
 		pq_beginmessage(&buf, 'S');
 		pq_sendint(&buf, TXN_START_PREPARED_RESULT, 4);
@@ -2412,7 +2428,7 @@ ProcessPrepareTransactionCommand(Port *myport, StringInfo message, bool is_backu
 
 	MemoryContextSwitchTo(oldContext);
 
-	elog(LOG, "Preparing transaction id %u", gxid);
+	elog(DEBUG1, "Preparing transaction id %u", gxid);
 
 	if (!is_backup)
 	{
@@ -2422,7 +2438,7 @@ ProcessPrepareTransactionCommand(Port *myport, StringInfo message, bool is_backu
 			GTM_Conn *oldconn = GetMyThreadInfo->thr_conn->standby;
 			int count = 0;
 
-			elog(LOG, "calling prepare_transaction() for standby GTM %p.", GetMyThreadInfo->thr_conn->standby);
+			elog(DEBUG1, "calling prepare_transaction() for standby GTM %p.", GetMyThreadInfo->thr_conn->standby);
 
 		retry:
 			bkup_prepare_transaction(GetMyThreadInfo->thr_conn->standby, gxid);
@@ -2434,7 +2450,7 @@ ProcessPrepareTransactionCommand(Port *myport, StringInfo message, bool is_backu
 			if (Backup_synchronously && (myport->remote_type != GTM_NODE_GTM_PROXY))
 				gtm_sync_standby(GetMyThreadInfo->thr_conn->standby);
 
-			elog(LOG, "prepare_transaction() GXID=%d done.", gxid);
+			elog(DEBUG1, "prepare_transaction() GXID=%d done.", gxid);
 		}
 		/* Respond to the client */
 		pq_beginmessage(&buf, 'S');
@@ -2545,6 +2561,7 @@ ProcessGetNextGXIDTransactionCommand(Port *myport, StringInfo message)
 	 */
 	GTM_RWLockAcquire(&GTMTransactions.gt_XidGenLock, GTM_LOCKMODE_WRITE);
 	next_gxid = GTMTransactions.gt_nextXid;
+
 	GTM_RWLockRelease(&GTMTransactions.gt_XidGenLock);
 
 	MemoryContextSwitchTo(oldContext);
@@ -2581,13 +2598,13 @@ GTM_SetShuttingDown(void)
 }
 
 void
-GTM_RestoreTxnInfo(int ctlfd, GlobalTransactionId next_gxid)
+GTM_RestoreTxnInfo(FILE *ctlf, GlobalTransactionId next_gxid)
 {
 	GlobalTransactionId saved_gxid;
 
-	if (ctlfd != -1)
+	if (ctlf)
 	{
-		if ((read(ctlfd, &saved_gxid, sizeof (saved_gxid)) != sizeof (saved_gxid)) &&
+		if ((fscanf(ctlf, "%u", &saved_gxid) != 1) &&
 			(!GlobalTransactionIdIsValid(next_gxid)))
 			next_gxid = InitialGXIDValue_Default;
 		else if (!GlobalTransactionIdIsValid(next_gxid))
@@ -2596,7 +2613,7 @@ GTM_RestoreTxnInfo(int ctlfd, GlobalTransactionId next_gxid)
 	else if (!GlobalTransactionIdIsValid(next_gxid))
 		next_gxid = InitialGXIDValue_Default;
 
-	elog(LOG, "Restoring last GXID to %u\n", next_gxid);
+	elog(DEBUG1, "Restoring last GXID to %u\n", next_gxid);
 
 	if (GlobalTransactionIdIsValid(next_gxid))
 		SetNextGlobalTransactionId(next_gxid);
@@ -2606,7 +2623,7 @@ GTM_RestoreTxnInfo(int ctlfd, GlobalTransactionId next_gxid)
 }
 
 void
-GTM_SaveTxnInfo(int ctlfd)
+GTM_SaveTxnInfo(FILE *ctlf)
 {
 	GlobalTransactionId next_gxid;
 
@@ -2614,8 +2631,25 @@ GTM_SaveTxnInfo(int ctlfd)
 
 	elog(LOG, "Saving transaction info - next_gxid: %u", next_gxid);
 
-	write(ctlfd, &next_gxid, sizeof (next_gxid));
+	fprintf(ctlf, "%u\n", next_gxid);
 }
+
+bool GTM_NeedXidRestoreUpdate(void)
+{
+	return(GlobalTransactionIdPrecedesOrEquals(GTMTransactions.gt_backedUpXid, GTMTransactions.gt_nextXid));
+}
+
+void GTM_WriteRestorePointXid(FILE *f)
+{
+	if ((MaxGlobalTransactionId - GTMTransactions.gt_nextXid) <= RestoreDuration)
+		GTMTransactions.gt_backedUpXid = GTMTransactions.gt_nextXid + RestoreDuration;
+	else
+		GTMTransactions.gt_backedUpXid = FirstNormalGlobalTransactionId + (RestoreDuration - (MaxGlobalTransactionId - GTMTransactions.gt_nextXid));
+	
+	elog(LOG, "Saving transaction restoration info, backed-up gxid: %u", GTMTransactions.gt_backedUpXid);
+	fprintf(f, "%u\n", GTMTransactions.gt_backedUpXid);
+}
+
 /*
  * TODO
  */
@@ -2625,3 +2659,4 @@ int GTM_GetAllTransactions(GTM_TransactionInfo txninfo[], uint32 txncnt);
  * TODO
  */
 uint32 GTM_GetAllPrepared(GlobalTransactionId gxids[], uint32 gxidcnt);
+
