@@ -4,7 +4,7 @@
  *
  * Portions Copyright (c) 1996-2009, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
- * Portions Copyright (c) 2010-2013, Postgres-XC Development Group
+ * Portions Copyright (c) 2010-2012 Postgres-XC Development Group
  *
  *
  * IDENTIFICATION
@@ -46,6 +46,7 @@
 #include "gtm/gtm_msg.h"
 #include "gtm/gtm_opt.h"
 #include "gtm/gtm_utils.h"
+#include "gtm/gtm_backup.h"
 
 extern int	optind;
 extern char *optarg;
@@ -115,6 +116,7 @@ static void checkDataDir(void);
 static void DeleteLockFile(const char *filename);
 static void PromoteToActive(void);
 static void ProcessSyncStandbyCommand(Port *myport, GTM_MessageType mtype, StringInfo message);
+static void ProcessBarrierCommand(Port *myport, GTM_MessageType mtype, StringInfo message);
 
 /*
  * One-time initialization. It's called immediately after the main process
@@ -128,9 +130,10 @@ MainThreadInit()
 	pthread_key_create(&threadinfo_key, NULL);
 
 	/*
-	 * Initialize the lock protecting the global threads info
+	 * Initialize the lock protecting the global threads info and backup lock info.
 	 */
 	GTM_RWLockInit(&GTMThreads->gt_lock);
+	GTM_RWLockInit(&gtm_bkup_lock);
 
 	/*
 	 * We are called even before memory context management is setup. We must
@@ -196,6 +199,7 @@ BaseInit()
 	GTM_InitTxnManager();
 	GTM_InitSeqManager();
 
+
 	/*
 	 * The memory context is now set up.
 	 * Add the thrinfo structure in the global array
@@ -288,7 +292,7 @@ main(int argc, char *argv[])
 	int			status;
 	int			i;
 	GlobalTransactionId next_gxid = InvalidGlobalTransactionId;
-	int			ctlfd;
+	FILE	   *ctlf;
 
 	/*
 	 * Local variable to hold command line options.
@@ -556,7 +560,7 @@ main(int argc, char *argv[])
 		elog(LOG, "Startup connection established with active-GTM.");
 	}
 
-	elog(LOG, "Starting GTM server at (%s:%d) -- control file %s", ListenAddresses, GTMPortNumber, GTMControlFile);
+	elog(DEBUG1, "Starting GTM server at (%s:%d) -- control file %s", ListenAddresses, GTMPortNumber, GTMControlFile);
 
 	/*
 	 * Read the last GXID and start from there
@@ -573,30 +577,36 @@ main(int argc, char *argv[])
 			elog(ERROR, "Failed to restore next/last gxid from the active-GTM.");
 			exit(1);
 		}
-		elog(LOG, "Restoring next/last gxid from the active-GTM succeeded.");
+		elog(DEBUG1, "Restoring next/last gxid from the active-GTM succeeded.");
 
 		if (!gtm_standby_restore_gxid())
 		{
 			elog(ERROR, "Failed to restore all of gxid(s) from the active-GTM.");
 			exit(1);
 		}
-		elog(LOG, "Restoring all of gxid(s) from the active-GTM succeeded.");
+		elog(DEBUG1, "Restoring all of gxid(s) from the active-GTM succeeded.");
 
 		if (!gtm_standby_restore_sequence())
 		{
 			elog(ERROR, "Failed to restore sequences from the active-GTM.");
 			exit(1);
 		}
-		elog(LOG, "Restoring sequences from the active-GTM succeeded.");
+		elog(DEBUG1, "Restoring sequences from the active-GTM succeeded.");
 	}
 	else
 	{
-		ctlfd = open(GTMControlFile, O_RDONLY);
-		GTM_RestoreTxnInfo(ctlfd, next_gxid);
-		GTM_RestoreSeqInfo(ctlfd);
-		close(ctlfd);
+		ctlf = fopen(GTMControlFile, "r");
+		GTM_RestoreTxnInfo(ctlf, next_gxid);
+		GTM_RestoreSeqInfo(ctlf);
+		if (ctlf)
+			fclose(ctlf);
 	}
 
+	/* Backup the restore point */
+
+	GTM_SetNeedBackup();
+	GTM_WriteRestorePoint();
+
 	if (Recovery_IsStandby())
 	{
 		if (!gtm_standby_register_self(NodeName, GTMPortNumber, GTMDataDir))
@@ -604,7 +614,7 @@ main(int argc, char *argv[])
 			elog(ERROR, "Failed to register myself on the active-GTM as a GTM node.");
 			exit(1);
 		}
-		elog(LOG, "Registering myself to the active-GTM as a GTM node succeeded.");
+		elog(DEBUG1, "Registering myself to the active-GTM as a GTM node succeeded.");
 	}
 
 	/* Recover Data of Registered nodes. */
@@ -615,7 +625,7 @@ main(int argc, char *argv[])
 			elog(ERROR, "Failed to restore node information from the active-GTM.");
 			exit(1);
 		}
-		elog(LOG, "Restoring node information from the active-GTM succeeded.");
+		elog(DEBUG1, "Restoring node information from the active-GTM succeeded.");
 
 		if (!gtm_standby_end_backup())
 		{
@@ -691,7 +701,7 @@ main(int argc, char *argv[])
 			elog(ERROR, "Failed to update the standby-GTM status as \"CONNECTED\".");
 			exit(1);
 		}
-		elog(LOG, "Updating the standby-GTM status as \"CONNECTED\" succeeded.");
+		elog(DEBUG1, "Updating the standby-GTM status as \"CONNECTED\" succeeded.");
 		if (!gtm_standby_finish_startup())
 		{
 			elog(ERROR, "Failed to close the initial connection to the active-GTM.");
@@ -781,7 +791,7 @@ ServerLoop(void)
 
 		if (GTMAbortPending)
 		{
-			int ctlfd;
+			FILE *ctlf;
 
 			/*
 			 * XXX We should do a clean shutdown here. For the time being, just
@@ -796,16 +806,15 @@ ServerLoop(void)
 			 */
 			GTM_SetShuttingDown();
 
-			ctlfd = open(GTMControlFile, O_WRONLY | O_TRUNC | O_CREAT,
-						 S_IRUSR | S_IWUSR);
-			if (ctlfd == -1)
+			ctlf = fopen(GTMControlFile, "w");
+			if (ctlf == NULL)
 			{
 				fprintf(stderr, "Failed to create/open the control file\n");
 				exit(2);
 			}
 
-			GTM_SaveTxnInfo(ctlfd);
-			GTM_SaveSeqInfo(ctlfd);
+			GTM_SaveTxnInfo(ctlf);
+			GTM_SaveSeqInfo(ctlf);
 
 #if 0
 			/*
@@ -821,7 +830,7 @@ ServerLoop(void)
 			}
 #endif
 
-			close(ctlfd);
+			fclose(ctlf);
 
 			exit(1);
 		}
@@ -1220,6 +1229,7 @@ ProcessCommand(Port *myport, StringInfo input_message)
 			break;
 		case MSG_END_BACKUP:
 			ProcessGTMEndBackup(myport, input_message);
+			break;
 		case MSG_NODE_BEGIN_REPLICATION_INIT:
 		case MSG_NODE_END_REPLICATION_INIT:
 		case MSG_TXN_BEGIN:
@@ -1282,9 +1292,13 @@ ProcessCommand(Port *myport, StringInfo input_message)
 			ProcessQueryCommand(myport, mtype, input_message);
 			break;
 
+		case MSG_BARRIER:
+		case MSG_BKUP_BARRIER:
+			ProcessBarrierCommand(myport, mtype, input_message);
+			break;
+
 		case MSG_BACKEND_DISCONNECT:
 			GTM_RemoveAllTransInfos(proxyhdr.ph_conid);
-
 			/* Mark PGXC Node as disconnected if backend disconnected is postmaster */
 			ProcessPGXCNodeBackendDisconnect(myport, input_message);
 			break;
@@ -1295,6 +1309,8 @@ ProcessCommand(Port *myport, StringInfo input_message)
 					 errmsg("invalid frontend message type %d",
 							mtype)));
 	}
+	if (GTM_NeedBackup())
+		GTM_WriteRestorePoint();
 }
 
 static int
@@ -1483,6 +1499,7 @@ ProcessTransactionCommand(Port *myport, GTM_MessageType mtype, StringInfo messag
 
 		case MSG_BKUP_TXN_BEGIN_GETGXID:
 			ProcessBkupBeginTransactionGetGXIDCommand(myport, message);
+			break;
 
 		case MSG_TXN_BEGIN_GETGXID_AUTOVACUUM:
 			ProcessBeginTransactionGetGXIDAutovacuumCommand(myport, message);
@@ -2044,7 +2061,7 @@ PromoteToActive(void)
 	 * Update the GTM config file for the next restart..
 	 */
 	conf_file = GetConfigOption("config_file", true);
-	elog(LOG, "Config file is %s...", conf_file);
+	elog(DEBUG1, "Config file is %s...", conf_file);
 	if ((fp = fopen(conf_file, PG_BINARY_A)) == NULL)
 	{
 		ereport(FATAL,
@@ -2062,12 +2079,71 @@ PromoteToActive(void)
 					"%Y-%m-%d %H:%M:%S %Z",
 					localtime(&stamp_time));
 
-		fprintf(fp, "#===================================================\n# Updated due to GTM promote request\n# %s\nstartup = ACT\n#===================================================\n", strfbuf);
+		fprintf(fp, 
+				"#===================================================\n"
+				"# Updated due to GTM promote request\n"
+				"# %s\nstartup = ACT\n"
+				"#===================================================\n", strfbuf);
 		if (fclose(fp))
 			ereport(FATAL,
 					(EINVAL,
 					 errmsg("could not close GTM configuration file \"%s\": %m",
 							conf_file)));
 	}
+	GTM_SetNeedBackup();
+	GTM_WriteRestorePoint();
 	return;
 }
+
+static void ProcessBarrierCommand(Port *myport, GTM_MessageType mtype, StringInfo message)
+{
+	int barrier_id_len;
+	char *barrier_id;
+	int count = 0;
+	GTM_Conn *oldconn = GetMyThreadInfo->thr_conn->standby;
+	StringInfoData buf;
+
+	barrier_id_len = pq_getmsgint(message, sizeof(int));
+	barrier_id = (char *)pq_getmsgbytes(message, barrier_id_len);
+	pq_getmsgend(message);
+
+	elog(INFO, "Processing BARRIER %s", barrier_id);
+
+	if ((mtype == MSG_BARRIER) && GetMyThreadInfo->thr_conn->standby)
+	{
+	retry:
+		bkup_report_barrier(GetMyThreadInfo->thr_conn->standby, barrier_id);
+		if (gtm_standby_check_communication_error(&count, oldconn))
+			goto retry;
+
+		if (Backup_synchronously && (myport->remote_type != GTM_NODE_GTM_PROXY))
+			gtm_sync_standby(GetMyThreadInfo->thr_conn->standby);
+	}
+
+	GTM_WriteBarrierBackup(barrier_id);
+
+	if (mtype == MSG_BARRIER)
+	{
+		/*
+		 * Send a SUCCESS message back to the client
+		 */
+		pq_beginmessage(&buf, 'S');
+		pq_sendint(&buf, BARRIER_RESULT, 4);
+		if (myport->remote_type == GTM_NODE_GTM_PROXY)
+		{
+			GTM_ProxyMsgHeader proxyhdr;
+			proxyhdr.ph_conid = myport->conn_id;
+			pq_sendbytes(&buf, (char *)&proxyhdr, sizeof (GTM_ProxyMsgHeader));
+		}
+		pq_endmessage(myport, &buf);
+
+		if (myport->remote_type != GTM_NODE_GTM_PROXY)
+		{
+			/* Flush standby first */
+			if (GetMyThreadInfo->thr_conn->standby)
+				gtmpqFlush(GetMyThreadInfo->thr_conn->standby);
+			pq_flush(myport);
+		}
+	}
+}
+	
