@@ -5,7 +5,7 @@
  *
  * Portions Copyright (c) 1996-2009, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
- * Portions Copyright (c) 2010-2013, Postgres-XC Development Group
+ * Portions Copyright (c) 2010-2012 Postgres-XC Development Group
  *
  *
  * IDENTIFICATION
@@ -120,7 +120,7 @@ pgxcnode_find_by_type(GTM_PGXCNodeType type, GTM_PGXCNodeInfo **data, size_t max
 			if (cur != NULL && cur->type == type)
 			{
 				data[node] = cur;
-				elog(LOG, "pgxcnode_find_by_type: cur=%p, ipaddress=%s", cur, cur->ipaddress);
+				elog(DEBUG1, "pgxcnode_find_by_type: cur=%p, ipaddress=%s", cur, cur->ipaddress);
 				node++;
 			}
 
@@ -394,10 +394,10 @@ Recovery_PGXCNodeRegister(GTM_PGXCNodeType	type,
 	nodeinfo->status = status;
 	nodeinfo->socket = socket;
 
-	elog(LOG, "Recovery_PGXCNodeRegister Request info: type=%d, nodename=%s, port=%d," \
+	elog(DEBUG1, "Recovery_PGXCNodeRegister Request info: type=%d, nodename=%s, port=%d," \
 			  "datafolder=%s, ipaddress=%s, status=%d",
 			  type, nodename, port, datafolder, ipaddress, status);
-	elog(LOG, "Recovery_PGXCNodeRegister Node info: type=%d, nodename=%s, port=%d, "\
+	elog(DEBUG1, "Recovery_PGXCNodeRegister Node info: type=%d, nodename=%s, port=%d, "\
 			  "datafolder=%s, ipaddress=%s, status=%d",
 			  nodeinfo->type, nodeinfo->nodename, nodeinfo->port,
 			  nodeinfo->datafolder, nodeinfo->ipaddress, nodeinfo->status);
@@ -417,7 +417,6 @@ Recovery_PGXCNodeRegister(GTM_PGXCNodeType	type,
 
 
 
-
 /*
  * Called at GTM shutdown, rewrite on disk register information
  * and write only data of nodes currently registered.
@@ -425,6 +424,7 @@ Recovery_PGXCNodeRegister(GTM_PGXCNodeType	type,
 void
 Recovery_SaveRegisterInfo(void)
 {
+#define Write(x, y, z) do{if (write(x, y, z) != (z)) elog(WARNING, "Write error in Recovery_SaveRegisterInfo");}while(0)
 	GTM_PGXCNodeInfoHashBucket *bucket;
 	gtm_ListCell *elem;
 	GTM_PGXCNodeInfo *nodeinfo = NULL;
@@ -462,62 +462,62 @@ Recovery_SaveRegisterInfo(void)
 
 			GTM_RWLockAcquire(&nodeinfo->node_lock, GTM_LOCKMODE_READ);
 
-			write(ctlfd, &NodeRegisterMagic, sizeof (NodeRegisterMagic));
+			Write(ctlfd, &NodeRegisterMagic, sizeof (NodeRegisterMagic));
 
-			write(ctlfd, &nodeinfo->type, sizeof (GTM_PGXCNodeType));
+			Write(ctlfd, &nodeinfo->type, sizeof (GTM_PGXCNodeType));
 			if (nodeinfo->nodename)
 			{
 				len = strlen(nodeinfo->nodename);
-				write(ctlfd, &len, sizeof(uint32));
-				write(ctlfd, nodeinfo->nodename, len);
+				Write(ctlfd, &len, sizeof(uint32));
+				Write(ctlfd, nodeinfo->nodename, len);
 			}
 			else
 			{
 				len = 0;
-				write(ctlfd, &len, sizeof(uint32));
+				Write(ctlfd, &len, sizeof(uint32));
 			}
 
-			write(ctlfd, &nodeinfo->port, sizeof (GTM_PGXCNodePort));
+			Write(ctlfd, &nodeinfo->port, sizeof (GTM_PGXCNodePort));
 
 			if (nodeinfo->proxyname)
 			{
 				len = strlen(nodeinfo->proxyname);
-				write(ctlfd, &len, sizeof(uint32));
-				write(ctlfd, nodeinfo->proxyname, len);
+				Write(ctlfd, &len, sizeof(uint32));
+				Write(ctlfd, nodeinfo->proxyname, len);
 			}
 			else
 			{
 				len = 0;
-				write(ctlfd, &len, sizeof(uint32));
+				Write(ctlfd, &len, sizeof(uint32));
 			}
 
-			write(ctlfd, &nodeinfo->status, sizeof (GTM_PGXCNodeStatus));
+			Write(ctlfd, &nodeinfo->status, sizeof (GTM_PGXCNodeStatus));
 
 			if (nodeinfo->ipaddress)
 			{
 				len = strlen(nodeinfo->ipaddress);
-				write(ctlfd, &len, sizeof(uint32));
-				write(ctlfd, nodeinfo->ipaddress, len);
+				Write(ctlfd, &len, sizeof(uint32));
+				Write(ctlfd, nodeinfo->ipaddress, len);
 			}
 			else
 			{
 				len = 0;
-				write(ctlfd, &len, sizeof(uint32));
+				Write(ctlfd, &len, sizeof(uint32));
 			}
 
 			if (nodeinfo->datafolder)
 			{
 				len = strlen(nodeinfo->datafolder);
-				write(ctlfd, &len, sizeof(uint32));
-				write(ctlfd, nodeinfo->datafolder, len);
+				Write(ctlfd, &len, sizeof(uint32));
+				Write(ctlfd, nodeinfo->datafolder, len);
 			}
 			else
 			{
 				len = 0;
-				write(ctlfd, &len, sizeof(uint32));
+				Write(ctlfd, &len, sizeof(uint32));
 			}
 
-			write(ctlfd, &NodeEndMagic, sizeof(NodeEndMagic));
+			Write(ctlfd, &NodeEndMagic, sizeof(NodeEndMagic));
 
 			GTM_RWLockRelease(&nodeinfo->node_lock);
 		}
@@ -534,6 +534,7 @@ Recovery_SaveRegisterInfo(void)
 	}
 
 	GTM_RWLockRelease(&RegisterFileLock);
+#undef Write
 }
 
 /*
@@ -542,6 +543,7 @@ Recovery_SaveRegisterInfo(void)
 void
 Recovery_RecordRegisterInfo(GTM_PGXCNodeInfo *nodeinfo, bool is_register)
 {
+#define Write(x, y, z) do{if (write(x, y, z) != (z)) elog(WARNING, "Write error in Recovery_RecordRegisterInfo");}while(0)
 	int ctlfd;
 	int len;
 
@@ -559,75 +561,76 @@ Recovery_RecordRegisterInfo(GTM_PGXCNodeInfo *nodeinfo, bool is_register)
 	GTM_RWLockAcquire(&nodeinfo->node_lock, GTM_LOCKMODE_READ);
 
 	if (is_register)
-		write(ctlfd, &NodeRegisterMagic, sizeof (NodeRegisterMagic));
+		Write(ctlfd, &NodeRegisterMagic, sizeof (NodeRegisterMagic));
 	else
-		write(ctlfd, &NodeUnregisterMagic, sizeof (NodeUnregisterMagic));
+		Write(ctlfd, &NodeUnregisterMagic, sizeof (NodeUnregisterMagic));
 
-	write(ctlfd, &nodeinfo->type, sizeof (GTM_PGXCNodeType));
+	Write(ctlfd, &nodeinfo->type, sizeof (GTM_PGXCNodeType));
 
 	if (nodeinfo->nodename)
 	{
 		len = strlen(nodeinfo->nodename);
-		write(ctlfd, &len, sizeof(uint32));
-		write(ctlfd, nodeinfo->nodename, len);
+		Write(ctlfd, &len, sizeof(uint32));
+		Write(ctlfd, nodeinfo->nodename, len);
 	}
 	else
 	{
 		len = 0;
-		write(ctlfd, &len, sizeof(uint32));
+		Write(ctlfd, &len, sizeof(uint32));
 	}
 
 	if (is_register)
 	{
 		int len;
 
-		write(ctlfd, &nodeinfo->port, sizeof (GTM_PGXCNodePort));
+		Write(ctlfd, &nodeinfo->port, sizeof (GTM_PGXCNodePort));
 
 		if (nodeinfo->proxyname)
 		{
 			len = strlen(nodeinfo->proxyname);
-			write(ctlfd, &len, sizeof(uint32));
-			write(ctlfd, nodeinfo->proxyname, len);
+			Write(ctlfd, &len, sizeof(uint32));
+			Write(ctlfd, nodeinfo->proxyname, len);
 		}
 		else
 		{
 			len = 0;
-			write(ctlfd, &len, sizeof(uint32));
+			Write(ctlfd, &len, sizeof(uint32));
 		}
 
-		write(ctlfd, &nodeinfo->status, sizeof (GTM_PGXCNodeStatus));
+		Write(ctlfd, &nodeinfo->status, sizeof (GTM_PGXCNodeStatus));
 
 		if (nodeinfo->ipaddress)
 		{
 			len = strlen(nodeinfo->ipaddress);
-			write(ctlfd, &len, sizeof(uint32));
-			write(ctlfd, nodeinfo->ipaddress, len);
+			Write(ctlfd, &len, sizeof(uint32));
+			Write(ctlfd, nodeinfo->ipaddress, len);
 		}
 		else
 		{
 			len = 0;
-			write(ctlfd, &len, sizeof(uint32));
+			Write(ctlfd, &len, sizeof(uint32));
 		}
 
 		if (nodeinfo->datafolder)
 		{
 			len = strlen(nodeinfo->datafolder);
-			write(ctlfd, &len, sizeof(uint32));
-			write(ctlfd, nodeinfo->datafolder, len);
+			Write(ctlfd, &len, sizeof(uint32));
+			Write(ctlfd, nodeinfo->datafolder, len);
 		}
 		else
 		{
 			len = 0;
-			write(ctlfd, &len, sizeof(uint32));
+			Write(ctlfd, &len, sizeof(uint32));
 		}
 	}
 
-	write(ctlfd, &NodeEndMagic, sizeof(NodeEndMagic));
+	Write(ctlfd, &NodeEndMagic, sizeof(NodeEndMagic));
 
 	GTM_RWLockRelease(&nodeinfo->node_lock);
 
 	close(ctlfd);
 	GTM_RWLockRelease(&RegisterFileLock);
+#undef Write
 }
 
 void
@@ -636,6 +639,8 @@ Recovery_RestoreRegisterInfo(void)
 	int magic;
 	int ctlfd;
 
+#define Read(x, y, z) do{if (read(x, y, z) != (z)) elog(WARNING, "Read error in Recovery_RestoreRegisterInfo");}while(0)
+
 	/* This is made when GTM/Proxy restarts, so it is not necessary to take a lock */
 	ctlfd = open(GTMPGXCNodeFile, O_RDONLY);
 
@@ -656,32 +661,32 @@ Recovery_RestoreRegisterInfo(void)
 			break;
 		}
 
-		read(ctlfd, &type, sizeof (GTM_PGXCNodeType));
+		Read(ctlfd, &type, sizeof (GTM_PGXCNodeType));
 		/* Read size of nodename string */
-		read(ctlfd, &len, sizeof (uint32));
+		Read(ctlfd, &len, sizeof (uint32));
 		nodename = (char *) palloc(len);
-		read(ctlfd, nodename, len);
+		Read(ctlfd, nodename, len);
 
 		if (magic == NodeRegisterMagic)
 		{
-			read(ctlfd, &port, sizeof (GTM_PGXCNodePort));
+			Read(ctlfd, &port, sizeof (GTM_PGXCNodePort));
 
 			/* Read size of proxyname string */
-			read(ctlfd, &len, sizeof (uint32));
+			Read(ctlfd, &len, sizeof (uint32));
 			proxyname = (char *) palloc(len);
-			read(ctlfd, proxyname, len);
+			Read(ctlfd, proxyname, len);
 
-			read(ctlfd, &status, sizeof (GTM_PGXCNodeStatus));
+			Read(ctlfd, &status, sizeof (GTM_PGXCNodeStatus));
 
 			/* Read size of ipaddress string */
-			read(ctlfd, &len, sizeof (uint32));
+			Read(ctlfd, &len, sizeof (uint32));
 			ipaddress = (char *) palloc(len);
-			read(ctlfd, ipaddress, len);
+			Read(ctlfd, ipaddress, len);
 
 			/* Read size of datafolder string */
-			read(ctlfd, &len, sizeof (uint32));
+			Read(ctlfd, &len, sizeof (uint32));
 			datafolder = (char *) palloc(len);
-			read(ctlfd, datafolder, len);
+			Read(ctlfd, datafolder, len);
 		}
 
 		/* Rebuild based on the records */
@@ -691,7 +696,7 @@ Recovery_RestoreRegisterInfo(void)
 		else
 			Recovery_PGXCNodeUnregister(type, nodename, true, 0);
 
-		read(ctlfd, &magic, sizeof(NodeEndMagic));
+		Read(ctlfd, &magic, sizeof(NodeEndMagic));
 
 		if (magic != NodeEndMagic)
 		{
@@ -701,6 +706,7 @@ Recovery_RestoreRegisterInfo(void)
 	}
 
 	close(ctlfd);
+#undef Read
 }
 
 void
@@ -847,7 +853,7 @@ ProcessPGXCNodeBackendDisconnect(Port *myport, StringInfo message)
 		GTM_Conn *oldconn = GetMyThreadInfo->thr_conn->standby;
 		int count = 0;
 
-		elog(LOG, "forwarding MSG_BACKEND_DISCONNECT to standby GTM %p.",
+		elog(DEBUG1, "forwarding MSG_BACKEND_DISCONNECT to standby GTM %p.",
 				  GetMyThreadInfo->thr_conn->standby);
 
 retry:
@@ -859,6 +865,6 @@ retry:
 		if (gtm_standby_check_communication_error(&count, oldconn))
 			goto retry;
 
-		elog(LOG, "MSG_BACKEND_DISCONNECT rc=%d done.", _rc);
+		elog(DEBUG1, "MSG_BACKEND_DISCONNECT rc=%d done.", _rc);
 	}
 }
