@@ -26,7 +26,7 @@ DECLARE
 		--Get all the node names
 		query_str_nodes := 'SELECT node_name FROM pgxc_node WHERE node_type = ''D''';
 		FOR row_name IN EXECUTE(query_str_nodes) LOOP
-			query_str := 'EXECUTE DIRECT ON ' || row_name.node_name || ' ''SELECT gid FROM pg_prepared_xact()''';
+			query_str := 'EXECUTE DIRECT ON (' || row_name.node_name || ') ''SELECT gid FROM pg_prepared_xact()''';
 			FOR row_data IN EXECUTE(query_str) LOOP
 				return next row_data.gid;
 			END LOOP;
@@ -37,4 +37,3 @@ LANGUAGE 'plpgsql';
 
 CREATE VIEW pgxc_prepared_xacts AS
     SELECT DISTINCT * from pgxc_prepared_xact();
-
